from ..definitions.columns.column import Column
from ..definitions.table import Table
class ChangeColumn:
    """ Generates a partial SQL command to change a column in a table """
    def __init__(self, column: Column):
        self.column = column

    def apply_to_table(self, table: Table):
        """ Changes the column in the table """
        table.change_column(self.column)

    def __str__(self):
        return 'CHANGE `%s` %s' % (self.column.name, str(self.column))



import unittest
from .numeric import Numeric
class NumericTest(unittest.TestCase):
    def test_is_the_same(self):
        attrs = {'length': '5,2', 'default': '0.00'}

        column1 = Numeric('test', 'DECIMAL', **attrs)
        column2 = Numeric('test', 'DECIMAL', **attrs)
        self.assertTrue(column1.is_really_the_same_as(column2))
        self.assertTrue(column2.is_really_the_same_as(column1))

        column1 = Numeric('test', 'DECIMAL', **attrs)
        column2 = Numeric('test', 'DECIMAL', **{**attrs, 'default': '0.02'})
        self.assertFalse(column1.is_really_the_same_as(column2))
        self.assertFalse(column2.is_really_the_same_as(column1))

        column1 = Numeric('test', 'DECIMAL', **attrs)
        column2 = Numeric('test', 'DECIMAL', **{**attrs, 'default': '0.001'})
        self.assertTrue(column1.is_really_the_same_as(column2))
        self.assertTrue(column2.is_really_the_same_as(column1))

        column1 = Numeric('test', 'DECIMAL', **attrs)
        column2 = Numeric('test', 'DECIMAL', **{**attrs, 'default': 0})
        self.assertTrue(column1.is_really_the_same_as(column2))
        self.assertTrue(column2.is_really_the_same_as(column1))

    def test_default_errors(self):
        self.assertEquals([f"Column 'test' of type 'DECIMAL' cannot have a string value as a default"],
                          Numeric('test', 'DECIMAL', default='').schema_errors)
        self.assertEquals([f"Column 'test' of type 'INT' must have an integer value as a default"],
                          Numeric('test', 'INT', default=5.0).schema_errors)
        self.assertEquals([f"Column 'test' of type 'BIT' must have a default of 1 or 0"],
                          Numeric('test', 'BIT', default=5).schema_errors)
        self.assertEquals([], Numeric('test', 'BIT', default=1).schema_errors)
        self.assertEquals([], Numeric('test', 'INT', default=5).schema_errors)
        self.assertEquals([], Numeric('test', 'FLOAT', default=5.0).schema_errors)

    def test_length_errors(self):
        self.assertEquals([f"Column 'test' of type 'FLOAT' cannot have a length"],
                          Numeric('test', 'FLOAT', length=5).schema_errors)
        self.assertEquals([f"Column 'test' of type 'INT' must have an integer value as its length"],
                          Numeric('test', 'INT', length='5,2').schema_errors)

    def test_character_set_errors(self):
        self.assertEquals([f"Column 'test' of type 'INT' cannot have a character set"],
                          Numeric('test', 'INT', character_set='UTF-8').schema_errors)
        self.assertEquals([f"Column 'test' of type 'INT' cannot have a collate"],
                          Numeric('test', 'INT', collate='UTF-8').schema_errors)

    def test_auto_increment_errors(self):
        self.assertEquals([f"Column 'test' of type 'FLOAT' cannot be an AUTO_INCREMENT"],
                          Numeric('test', 'FLOAT', auto_increment=True).schema_errors)
        self.assertEquals([], Numeric('test', 'FLOAT', auto_increment=False).schema_errors)
        self.assertEquals([], Numeric('test', 'INT', auto_increment=True).schema_errors)



from ..definitions.constraint import Constraint
from ..definitions.table import Table
class RemoveConstraint:
    """ Generates a partial SQL command to drop a foreign key from a table """
    def __init__(self, constraint: Constraint):
        self._constraint = constraint

    def apply_to_table(self, table: Table):
        """ Removes the constraint from the table """
        table.remove_constraint(self._constraint)

    def __str__(self):
        return 'DROP FOREIGN KEY `%s`' % (self._constraint.name)



import unittest
from .create_table import CreateTable
from ..definitions.columns import String, Numeric
from ..definitions.index import Index
from ..definitions.table import Table
from ..definitions.option import Option
class CreateTableTest(unittest.TestCase):
    def test_as_string(self):
        table = Table(
            'test_table', [
                Numeric('id', 'INT', length=10, unsigned=True, auto_increment=True),
                String('name', 'VARCHAR', length=255)
            ], [Index('id', ['id'], 'PRIMARY')], [], [Option('ENGINE', 'InnoDB')]
        )
        self.assertEquals(
            """CREATE TABLE `test_table` (
  `id` INT(10) UNSIGNED AUTO_INCREMENT,
  `name` VARCHAR(255),
  PRIMARY KEY `id` (`id`)
) ENGINE=InnoDB;""", str(CreateTable(table, True))
        )
        self.assertEquals(
            """CREATE TABLE `test_table` (`id` INT(10) UNSIGNED AUTO_INCREMENT,`name` VARCHAR(255),PRIMARY KEY `id` (`id`)) ENGINE=InnoDB;""",
            str(CreateTable(table, False))
        )



from .base import Base
from mygrations.formats.mysql.file_reader.database import Database as DatabaseParser
from mygrations.formats.mysql.mygrations.mygration import Mygration
def execute(options, print_results=True):
    obj = Check(options)
    return obj.execute(print_results=print_results)
class Check(Base):
    def execute(self, print_results=True):
        files_database = DatabaseParser(self.get_sql_files())

        # any errors or warnings?
        errors = []
        if files_database.errors:
            errors.append('Errors found in *.sql files:')
            errors.extend(files_database.errors)

        if files_database.warnings:
            errors.append('Warnings found in *.sql files:')
            errors.extend(files_database.warnings)

        if print_results:
            if not errors:
                print("No problems found")
            else:
                print("\n".join(errors))
        return (errors, True if not len(errors) else False)



from ..definitions.constraint import Constraint
from ..definitions.table import Table
class ChangeConstraint:
    """ Generates a partial SQL command to change a constraint in a table """
    def __init__(self, constraint: Constraint):
        self.constraint = constraint

    def apply_to_table(self, table: Table):
        """ Changes the constraint in the table """
        table.change_constraint(self.constraint)

    def __str__(self):
        return 'DROP FOREIGN KEY `%s`, ADD %s' % (self.constraint.name, str(self.constraint))



from typing import Union, List
class Option:
    _name: str = ''
    _value: str = ''
    _schema_errors: List[str] = None
    _schema_warnings: List[str] = None
    _parsing_errors: List[str] = None
    _parsing_warnings: List[str] = None

    def __init__(self, name: str = '', value: str = ''):
        self._name = name
        self._value = value
        self._schema_errors = None
        self._schema_warnings = None
        self._parsing_errors = None
        self._parsing_warnings = None

    @property
    def name(self) -> str:
        """ Public getter.  Returns the name of the option.

        :returns: The option name
        """
        return self._name

    @property
    def value(self) -> str:
        """ Public getter.  Returns the value of the option.

        :returns: The option value
        """
        return self._value

    @property
    def schema_errors(self) -> List[str]:
        """ Returns a list of schema errors """
        if self._schema_errors is None:
            self._check_for_schema_errors_and_warnings()
        return self._schema_errors

    @property
    def schema_warnings(self) -> List[str]:
        """ Returns a list of schema warnings """
        if self._schema_errors is None:
            self._check_for_schema_errors_and_warnings()
        return self._schema_warnings

    @property
    def parsing_errors(self) -> List[str]:
        """ Returns a list of parsing errors """
        return self._parsing_errors if self._parsing_errors is not None else []

    @property
    def parsing_warnings(self) -> List[str]:
        """ Returns a list of parsing/table warnings """
        return self._parsing_warnings if self._parsing_warnings is not None else []

    def _check_for_schema_errors_and_warnings(self):
        self._schema_errors = []
        self._schema_warnings = []
        for required in ['name', 'value']:
            if not getattr(self, required):
                self._schema_errors.append(f"Missing {required} for option {self.name}")
        if len(self.name) > 64:
            self._schema_errors.append('Option name %s must be <=64 characters long' % (self._name))



import unittest
from .remove_row import RemoveRow
class RemoveRowTest(unittest.TestCase):
    def test_as_string(self):
        self.assertEquals("DELETE FROM `test_table` WHERE id='10';", str(RemoveRow('test_table', 10)))



import unittest
from collections import OrderedDict
from .columns import Numeric, String
from .constraint import Constraint
from .index import Index
from .option import Option
from .table import Table
class TestTable(unittest.TestCase):
    def setUp(self):
        self.id = Numeric('id', 'INT', length=10, unsigned=True, null=False, auto_increment=True)
        self.user_id = Numeric('user_id', 'INT', length=10, unsigned=True, null=True)
        self.age = Numeric('age', 'INT', length=10, unsigned=True, default=0)
        self.name = String('name', 'VARCHAR', length=255, default='', null=False)
        self.id_index = Index('id_index', ['id'], 'PRIMARY')
        self.name_age_index = Index('name_age', ['name', 'age'], 'INDEX')
        self.user_id_index = Index('user_id_index', ['user_id'], 'INDEX')
        self.user_id_constraint = Constraint('user_id_fk', 'user_id', 'users', 'id', 'SET NULL', 'SET NULL')
        self.engine = Option('ENGINE', 'InnoDB')

    def test_can_create(self):
        table = Table(
            'registrations',
            [self.id, self.user_id, self.age, self.name],
            [self.id_index, self.name_age_index, self.user_id_index],
            [self.user_id_constraint],
            [self.engine],
        )

        self.assertEquals('registrations', table.name)
        self.assertEquals(['id', 'user_id', 'age', 'name'], [column.name for column in table.columns.values()])
        self.assertEquals(['user_id_fk'], [constraint.name for constraint in table.constraints.values()])
        self.assertEquals(['id_index', 'name_age', 'user_id_index'], [index.name for index in table.indexes.values()])
        self.assertEquals([self.engine], table.options)
        self.assertEquals(self.id_index, table.primary)
        self.assertEquals([], table.schema_errors)
        self.assertEquals([], table.schema_warnings)

    def test_require_basics(self):
        table = Table('', [self.id], [self.id_index], [], [])
        self.assertEquals(["Table missing name"], table.schema_errors)

        table = Table('no_cols', [], [], [], [])
        self.assertEquals(["Table 'no_cols' does not have any columns"], table.schema_errors)

    def test_include_child_errors(self):
        bad_column = String('text', 'TEXT', default='', has_default=True)
        bad_index = Index('no_cols', [], 'INDEX')
        bad_constraint = Constraint('user_id_fk', 'user_id', 'users', 'id', on_delete='CASCAD')
        table = Table(
            'errors',
            [self.id, self.user_id, bad_column],
            [self.id_index, bad_index],
            [bad_constraint],
            [self.engine],
        )
        self.assertEquals([
            "Column 'text' of type 'TEXT' cannot have a default in table 'errors'",
            "Missing columns for index no_cols in table 'errors'",
            "ON DELETE action of 'CASCAD' for constraint 'user_id_fk' is not a valid ON DELETE action in table 'errors'",
        ], table.schema_errors)

    def test_check_primaries(self):
        table = Table('errors', [self.id], [self.id_index, Index('id_primary', ['id'], 'PRIMARY')], [], [])
        self.assertEquals(["Table 'errors' has more than one PRIMARY index"], table.schema_errors)

        table = Table('errors', [self.id], [], [], [])
        self.assertEquals(["Table 'errors' has an AUTO_INCREMENT column but is missing the PRIMARY index"],
                          table.schema_errors)

        table = Table(
            'errors',
            [self.id, self.user_id],
            [Index('user_id', ['user_id'], 'PRIMARY')],
            [],
            [],
        )
        self.assertEquals([
            "Mismatched indexes in table 'errors': column 'id' is the AUTO_INCREMENT column but 'user_id' is the PRIMARY index column"
        ], table.schema_errors)

    def test_check_missing_index_columns(self):
        table = Table(
            'errors', [self.id, self.user_id],
            [self.id_index, Index('bad_index', ['non_column'], 'INDEX')], [], []
        )
        self.assertEquals(["Table 'errors' has index 'bad_index' that references non-existent column 'non_column'"],
                          table.schema_errors)



import unittest
from .change_column import ChangeColumn
from ..definitions.columns import String
class ChangeColumnTest(unittest.TestCase):
    def test_as_string(self):
        name = String('name', 'VARCHAR', length=255)
        self.assertEquals("CHANGE `name` `name` VARCHAR(255)", str(ChangeColumn(name)))



import unittest

from mygrations.formats.mysql.file_reader.database import Database
class DatabaseTest(unittest.TestCase):
    def test_simple(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """, """
            CREATE TABLE `more_logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `more_messages` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        database = Database(strings)

        # our parser should have a table!
        self.assertTrue('logs' in database.tables)
        self.assertTrue('more_logs' in database.tables)

    def test_with_rows(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT NOT NULL,
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """, """
            INSERT INTO logs (id,message,traceback) VALUES (1,'hi','sup'),(2,'what','kay');
        """
        ]
        database = Database(strings)

        # our parser should have a table!
        self.assertTrue('logs' in database.tables)

        rows = database.tables['logs'].rows
        self.assertEquals('hi', rows['1']['message'])
        self.assertEquals('sup', rows['1']['traceback'])
        self.assertEquals('what', rows['2']['message'])
        self.assertEquals('kay', rows['2']['traceback'])

    def test_keeps_errors(self):

        strings = [
            """
            CREATE TABLE `logs` (
                `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
                `message` TEXT DEFAULT '',
                `traceback` text,
                PRIMARY KEY (`id`)
            ) ENGINE=InnoDB DEFAULT CHARSET=utf8;
        """
        ]
        database = Database(strings)

        self.assertEquals(1, len(database.errors))
        self.assertTrue("Column 'message' of type 'TEXT' cannot have a default in table 'logs'" in database.errors[0])

    def test_check_mismatched_default_value_and_column(self):
        database = Database(
            """
CREATE TABLE `test_table` (
  `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
  `numeric_column` TINYINT(1) UNSIGNED NOT NULL DEFAULT '',
  PRIMARY KEY (`id`)
);
        """
        )

        self.assertTrue('test_table' in database.tables)
        self.assertEqual(1, len(database.errors))
        self.assertEqual(
            "Column 'numeric_column' of type 'TINYINT' cannot have a string value as a default in table 'test_table'",
            database.errors[0]
        )

    def test_duplicate_foreign_key_name(self):
        database = Database(
            """
CREATE TABLE `roles` (
  `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
  `people_id` INT(10) UNSIGNED NOT NULL,
  `name` varchar(255) NOT NULL DEFAULT '',
  PRIMARY KEY (`id`),
  KEY `role_people_id` (`people_id`),
  CONSTRAINT `people_role_id_fk` FOREIGN KEY (`people_id`) REFERENCES `people` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_general_ci;

CREATE TABLE `people` (
  `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
  `role_id` INT(10) UNSIGNED NOT NULL,
  `name` varchar(255) NOT NULL DEFAULT '',
  PRIMARY KEY (`id`),
  KEY `people_role_id` (`role_id`),
  CONSTRAINT `people_role_id_fk` FOREIGN KEY (`role_id`) REFERENCES `roles` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_general_ci;
        """
        )
        self.assertTrue('roles' in database.tables)
        self.assertTrue('people' in database.tables)
        self.assertEqual(1, len(database.errors))
        self.assertEqual(
            "Duplicate foreign key: foreign key named 'people_role_id_fk' exists in tables 'roles' and 'people'",
            database.errors[0]
        )

    def test_foreign_key_non_column(self):
        database = Database(
            """
CREATE TABLE `roles` (
  `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
  `name` varchar(255) NOT NULL DEFAULT '',
  PRIMARY KEY (`id`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_general_ci;

CREATE TABLE `people` (
  `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
  `role_id` INT(10) UNSIGNED NOT NULL,
  `name` varchar(255) NOT NULL DEFAULT '',
  PRIMARY KEY (`id`),
  KEY `people_role_id` (`role_id`),
  CONSTRAINT `people_role_id_fk` FOREIGN KEY (`roles_id`) REFERENCES `roles` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_general_ci;
        """
        )
        self.assertTrue('roles' in database.tables)
        self.assertTrue('people' in database.tables)
        self.assertEqual(1, len(database.errors))
        self.assertEqual(
            "Constraint error for foreign key `people_role_id_fk`: sets constraint on column `people`.`roles_id`, but this column does not exist",
            database.errors[0]
        )

    def test_missing_key_column(self):
        database = Database(
            """
CREATE TABLE `roles` (
  `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
  `name` VARCHAR(255) NOT NULL DEFAULT '',
  PRIMARY KEY (`id`),
  KEY `account_id_key` (`account_id`)
);
        """
        )

        self.assertTrue('roles' in database.tables)
        self.assertEqual(1, len(database.errors))
        self.assertEqual(
            "Table 'roles' has index 'account_id_key' that references non-existent column 'account_id'",
            database.errors[0]
        )

    def test_mismatched_column(self):
        database = Database(
            """
CREATE TABLE `roles` (
    `id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
    `name` VARCHAR(255) NOT NULL DEFAULT '',
    PRIMARY KEY (`id`)
);
INSERT INTO `roles` (`id`,`name`,`description`) VALUES (1,'asdf','more');
        """
        )

        self.assertEqual(1, len(database.errors))
        self.assertEquals(
            "Insert error: insert command attempts to set column 'description' for table 'roles' but the column does not exist in the table.",
            database.errors[0]
        )



from mygrations.core.parse.parser import Parser
from mygrations.formats.mysql.definitions.index import Index
class IndexKey(Parser, Index):
    _index_type = 'index'
    has_comma = False

    # KEY account_id (account_id,name)
    rules = [{
        'type': 'regexp',
        'name': 'index_name',
        'value': '(KEY)|(INDEX)'
    }, {
        'type': 'regexp',
        'name': 'name',
        'value': '[^\(\s\)]+'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type': 'delimited',
        'name': 'columns',
        'separator': ',',
        'quote': '`'
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def process(self):
        self._name = self._values['name'].strip().strip('`')
        self._columns = self._values['columns']
        self.has_comma = True if 'ending_comma' in self._values else False



from mygrations.core.parse.parser import Parser
from .type import Type
class TypeCharacter(Parser, Type):

    allowed_collation_types = {'char': True, 'varchar': True}

    # in this case we have much less disallowed than allowed
    disallowed_types = {
        'date': True,
        'year': True,
        'tinyblob': True,
        'blob': True,
        'mediumblob': True,
        'longblob': True,
        'tinytext': True,
        'text': True,
        'mediumtext': True,
        'longtext': True,
        'json': True
    }

    has_comma = False

    # name varchar(255) NOT NULL DEFAULT '' CHARACTER SET uf8 COLLATE utf8
    # hackish? maybe (i.e. yes) the repeated COLLATE and CHARACTER SETS take care of uncertain
    # ordering.  I could also use `children` which is order-agnostic, but I'm being lazy
    rules = [{
        'type': 'regexp',
        'value': '[^\(\s\)]+',
        'name': 'name'
    }, {
        'type': 'regexp',
        'value': '\w+',
        'name': 'type'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type': 'regexp',
        'value': '\d+',
        'name': 'length'
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'CHARACTER SET ([^\(\s\),]+)',
        'name': 'character_set',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'NOT NULL',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'CHARACTER SET ([^\(\s\),]+)',
        'name': 'character_set',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'DEFAULT ([^\(\s\),]+)',
        'optional': True,
        'name': 'default'
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'CHARACTER SET ([^\(\s\),]+)',
        'name': 'character_set',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'COLLATE ([^\(\s\),]+)',
        'name': 'collate',
        'optional': True
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def process(self):

        self.has_comma = True if 'ending_comma' in self._values else False

        self._parsing_errors = []
        self._parsing_warnings = []
        self._schema_errors = []
        self._schema_warnings = []
        self._name = self._values['name'].strip('`')
        self._column_type = self._values['type']
        self._length = self._values['length']
        self._null = False if 'NOT NULL' in self._values else True
        self._has_default = 'default' in self._values
        self._default = self._values['default'] if 'default' in self._values else None
        self._character_set = self._values['character_set'] if 'character_set' in self._values else None
        self._collate = self._values['collate'] if 'collate' in self._values else None

        # make sense of the default
        if self._default and len(self._default) >= 2 and self._default[0] == "'" and self._default[-1] == "'":
            self._default = self._default.strip("'")
        elif self._default:
            if self._default.lower() == 'null':
                self._default = None
            elif not self._default.isdigit():
                self._parsing_warnings.append(
                    'Default value of "%s" should have quotes for field %s' % (self._default, self._name)
                )

        if self._character_set and len(self._character_set
                                       ) >= 2 and self._character_set[0] == "'" and self._character_set[-1] == "'":
            self._character_set = self._character_set.strip("'")

        if self._collate and len(self._collate) >= 2 and self._collate[0] == "'" and self._collate[-1] == "'":
            self._collate = self._collate.strip("'")

        self._attributes = {}
        if self._character_set:
            self._attributes['CHARACTER SET'] = self._character_set
        if self._collate:
            self._attributes['COLLATE'] = self._collate



import re
from .rule_children import RuleChildren
from .rule_delimited import RuleDelimited
from .rule_literal import RuleLiteral
from .rule_regexp import RuleRegexp
class Parser:
    rule_types = {'children': RuleChildren, 'delimited': RuleDelimited, 'literal': RuleLiteral, 'regexp': RuleRegexp}
    _parsing_errors = None
    _parsing_warnings = None
    _values = None

    @property
    def parsing_errors(self):
        """ Public getter.  Returns a list of parsing errors

        :returns: A list of parsing errors
        :rtype: list
        """
        return self._parsing_errors if self._parsing_errors is not None else []

    @property
    def parsing_warnings(self):
        """ Public getter.  Returns a list of parsing warnings

        :returns: A list of parsing warnings
        :rtype: list
        """
        return self._parsing_warnings if self._parsing_warnings is not None else []

    def __getitem__(self, key):

        return self._values[key]

    def __len__(self):

        return len(self._values)

    def has_key(self, key):

        return key in self._values

    def keys(self):

        return self._values.keys()

    def values(self):

        return self._values.values()

    def items(self):

        return self._values.items()

    def get_rule_parser(self, rule, next_rule):

        rule_type = rule['type']

        # keeping this simple for now
        if not rule_type in self.rule_types:
            raise ValueError('Unknown rule type %s for class %s' % (rule_type, self.__class__))

        return self.rule_types[rule_type](self, rule, next_rule)

    def parse(self, string=''):

        # first thing first, some initial string cleaning.  Clean spaces
        # from start and end and replace any multi-spaces with a single space.
        string = re.sub('\s+', ' ', string).strip()
        number_rules = len(self.rules)
        self._values = {}

        for (rule_index, rule) in enumerate(self.rules):

            # do we have a next rule?
            next_rule = self.rules[rule_index + 1] if rule_index < number_rules - 1 else False

            # now we can parse
            rule_parser = self.get_rule_parser(rule, next_rule)

            # does it match?  Check for a lack of match and deal with that first
            if not rule_parser.parse(string):

                # if this rule wasn't optional then we just don't match
                if not 'optional' in rule or not rule['optional']:
                    self.matched = False
                    return string

                # otherwise just keep going
                continue

            # we did match!  Yeah!
            self._values[rule_parser.name] = rule_parser.result
            string = rule_parser.leftovers

            # we are all done if we have nothing left
            if not string:
                break

        # if we got here then we got to the end, but we may not be done.  If we have more required
        # rules left that haven't been matched, then we don't match.

        # did we check every required rule?
        for check_index in range(rule_index + 1, number_rules):
            rule = self.rules[check_index]
            if not 'optional' in rule or not rule['optional']:
                self.matched = False
                return string

        # if we got here then we didn't match everything, but we fulfilled all of our
        # required rules.  As a result, we are done!
        self.process()
        self.matched = True
        return string

    def process(self):
        """ parser.process()

        Processes the results of the parsing process.  Only called if a match is found.  No input
        or output as it modifies the parser object in place, populating attributes as needed.
        """

        pass



import unittest
from .add_index import AddIndex
from .alter_table import AlterTable
from ..definitions.index import Index
from ..definitions.table import Table
from ..definitions.columns import Numeric
class AlterTableTest(unittest.TestCase):
    def test_as_string(self):
        id_column = Numeric('id', 'INT', length=10, unsigned=True, null=False, auto_increment=True)
        user_id_column = Numeric('user_id', 'INT', length=10, unsigned=True, null=False)
        id_index = Index('id', ['id'], 'primary')
        user_id_index = Index('user_id', ['user_id', 'account_id'], 'index')

        alter_table = AlterTable('orders')
        alter_table.add_operation(AddIndex(id_index))
        alter_table.add_operation(AddIndex(user_id_index))

        self.assertEquals(
            "ALTER TABLE `orders` ADD PRIMARY KEY `id` (`id`), ADD KEY `user_id` (`user_id`,`account_id`);",
            str(alter_table)
        )



class CreateTable:
    """ Generates an SQL command to create a table """
    _table = None
    _nice: bool

    def __init__(self, table, nice=False):
        self._table = table
        self._nice = nice

    @property
    def table_name(self) -> str:
        """ Returns the name of the table. """
        return self._table.name

    def __str__(self):
        newline = '\n' if self._nice else ''
        padding = '  ' if self._nice else ''
        body_separator = ',\n' if self._nice else ','
        body = ['%s%s' % (padding, str(self._table.columns[col])) for col in self._table.columns]
        body.extend(['%s%s' % (padding, str(self._table.indexes[index])) for index in self._table.indexes])
        body.extend([
            '%s%s' % (padding, str(self._table.constraints[constraint])) for constraint in self._table.constraints
        ])
        options = ['%s=%s' % (opt.name, opt.value) for opt in self._table.options]
        if options:
            options = ' %s' % ' '.join(options)
        else:
            options = ''
        return 'CREATE TABLE `%s` (%s%s%s)%s;' % (
            self._table.name, newline, body_separator.join(body), newline, options
        )



from ..definitions.index import Index
from ..definitions.table import Table
class ChangeIndex:
    """ Generates a partial SQL command to change an index in a table """
    def __init__(self, index: Index):
        self.index = index

    def apply_to_table(self, table: Table):
        """ Changes the index in the table """
        table.change_index(self.index)

    def __str__(self):
        return 'DROP KEY `%s`, ADD %s' % (self.index.name, str(self.index))



import unittest
from .remove_column import RemoveColumn
from ..definitions.columns import String
class RemoveColumnTest(unittest.TestCase):
    def test_as_string(self):
        string = String('name', 'VARCHAR', length=255, default='', null=False)
        self.assertEquals("DROP `name`", str(RemoveColumn(string)))



from __future__ import annotations
from typing import Any, Dict, List, Set, Union
from collections import OrderedDict
from .rows import Rows
from .option import Option
from .columns.column import Column
from .index import Index
from .constraint import Constraint
from ..operations.create_table import CreateTable
class Table:
    _columns: Dict[str, Column] = None
    _constraints: Dict[str, Constraint] = None
    _indexes: Dict[str, Constraint] = None
    _indexed_columns = None
    _name: str = ''
    _options: List[Option] = None
    _primary: Index = None
    _rows: Dict[int, List[Union[str, int]]] = None
    _tracking_rows: bool = False
    _schema_errors: List[str] = None
    _schema_warnings: List[str] = None
    _global_errors: List[str] = None
    _global_warnings: List[str] = None

    def __init__(
        self,
        name: str = '',
        columns: List[Column] = None,
        indexes: List[Index] = None,
        constraints: List[Constraint] = None,
        options: List[Option] = None
    ):
        self._name = name
        self._options = [*options] if options else []
        self._primary = None
        self._constraints = {}
        self._columns = {}
        self._indexes = {}
        self._parsing_errors = None
        self._parsing_warnings = None
        self._schema_errors = None
        self._schema_warnings = None
        self._global_errors = []
        self._global_warnings = []
        self._final_global_errors = None
        self._final_global_warnings = None
        self._raw_columns = columns if columns is not None else []
        self._raw_constraints = constraints if constraints is not None else []
        self._raw_indexes = indexes if indexes is not None else []

        if columns is not None:
            for column in columns:
                self.add_column(column)
        if constraints is not None:
            for constraint in constraints:
                self.add_constraint(constraint)
        if indexes is not None:
            for index in indexes:
                self.add_index(index)

    # Errors are no longer tracked live. Instead, I need to finish adjusting the flow so that if any
    # of the error fetchers are called, then the errors are collected from all children and then the
    # proper errors are returned.

    @property
    def global_errors(self):
        """ Public getter.  Returns a list of schema errors

        :returns: A list of schema errors
        :rtype: list
        """
        if self._schema_errors is None:
            self._find_all_errors()
        return self._final_global_errors

    @property
    def global_warnings(self):
        """ Public getter.  Returns a list of schema warnings

        :returns: A list of schema warnings
        :rtype: list
        """
        if self._schema_errors is None:
            self._find_all_errors()
        return self._final_global_warnings

    @property
    def schema_errors(self):
        """ Public getter.  Returns a list of schema errors

        :returns: A list of schema errors
        :rtype: list
        """
        if self._schema_errors is None:
            self._find_all_errors()
        return self._schema_errors

    @property
    def schema_warnings(self):
        """ Public getter.  Returns a list of schema warnings

        :returns: A list of schema warnings
        :rtype: list
        """
        if self._schema_errors is None:
            self._find_all_errors()
        return self._schema_warnings

    @property
    def parsing_errors(self):
        """ Public getter.  Returns a list of schema errors

        :returns: A list of schema errors
        :rtype: list
        """
        if self._schema_errors is None:
            self._find_all_errors()
        return self._parsing_errors

    @property
    def parsing_warnings(self):
        """ Public getter.  Returns a list of schema warnings

        :returns: A list of schema warnings
        :rtype: list
        """
        if self._schema_errors is None:
            self._find_all_errors()
        return self._parsing_warnings

    @property
    def name(self):
        """ Public getter.  Returns the name of the table.

        :returns: The table name
        :rtype: string
        """

        return self._name

    @property
    def options(self):
        """ Public getter.  Returns a list of table options

        :returns: Table options
        :rtype: list
        """

        return self._options

    def _find_all_errors(self):
        self._schema_errors = []
        self._schema_warnings = []
        self._parsing_errors = []
        self._parsing_warnings = []
        if self._global_errors is not None:
            self._final_global_errors = [*self._global_errors]
        if self._global_warnings is not None:
            self._final_global_warnings = [*self._global_warnings]

        if not self.name:
            self._schema_errors.append("Table missing name")
        if not len(self.columns):
            self._schema_errors.append(f"Table '{self.name}' does not have any columns")

        # start with errors from "children" and append the table name for context
        for type_to_check in [self._columns, self._indexes, self._constraints]:
            for to_check in type_to_check.values():
                for error in to_check.schema_errors:
                    self._schema_errors.append(f"{error} in table '{self.name}'")
                for warning in to_check.schema_warnings:
                    self._schema_warnings.append(f"{warning} in table '{self.name}'")
                for error in to_check.parsing_errors:
                    self._parsing_errors.append(f"{error} in table '{self.name}'")
                for warning in to_check.parsing_warnings:
                    self._parsing_warnings.append(f"{error} in table '{self.name}'")
                if hasattr(to_check, 'global_errors'):
                    for warning in to_check.global_errors:
                        self._final_global_errors.append(f"{error} in table '{self.name}'")
                if hasattr(to_check, 'global_warnings'):
                    for warning in to_check.global_warnings:
                        self._final_global_warnings.append(f"{error} in table '{self.name}'")

        # duplicate names.  This shouldn't really be possible anymore because the add_ methods will
        # throw an exception if we try to add a duplicate, but I'll leave this in just in case.
        for (name, to_check) in {
            'columns': self._columns,
            'constraints': self._constraints,
            'indexes': self._indexes
        }.items():
            if len(to_check) == len(getattr(self, name)):
                continue
            label = name.rstrip('es').rstrip('s')
            found = {}
            duplicates = {}
            for item in to_check:
                if item.name in found:
                    duplicates[item.name] = True
                    continue
                found[item.name] = True
            for key in duplicates.keys():
                self._schema_errors.append(f"Duplicate {label} name found in table '{self.name}': '{key}'")

        # more than one primary key
        primaries = list(filter(lambda index: index.is_primary(), self._indexes.values()))
        if len(primaries) > 1:
            self._schema_errors.append(f"Table '{self.name}' has more than one PRIMARY index")

        # auto increment checks
        if self._columns:
            auto_increment = list(filter(lambda column: column.auto_increment, self._columns.values()))
            if len(auto_increment):
                if len(auto_increment) > 1:
                    self._schema_errors.append(f"Table '{self.name}' has more than one AUTO_INCREMENT column")
                elif not primaries:
                    self._schema_errors.append(
                        f"Table '{self.name}' has an AUTO_INCREMENT column but is missing the PRIMARY index"
                    )
                elif primaries[0].columns[0] != auto_increment[0].name:
                    self.schema_errors.append(
                        "Mismatched indexes in table '%s': column '%s' is the AUTO_INCREMENT column but '%s' is the PRIMARY index column"
                        % (self.name, auto_increment[0].name, primaries[0].columns[0])
                    )

        # indexes on non-existent columns
        for index in self._indexes.values():
            for column in index.columns:
                if not column in self.columns:
                    self._schema_errors.append(
                        f"Table '{self.name}' has index '{index.name}' that references non-existent column '{column}'"
                    )

        # we don't bother checking the constraints to see if they are valid because these are
        # best checked at the database level (since, by definition, foreign key constraints are *typically*
        # against other tables, not within a single table.

    def mark_tracking_rows(self):
        """ Marks the table as having had its rows read, for bookeeping purposes

        The reason this is here is because we keep track of whether or not we are "tracking" rows
        for the current table.  This is for bookeeping purposes, largely as a safe-guard attempt
        to try to identify any more subtle bugs that might creep in.  Normally self._tracking_rows
        gets set to True when we add rows to a table, but if a table is empty then rows will never
        get added, and instead this method must be called to mark the rows as "tracked".
        """
        self._tracking_rows = True
        if self._rows is None:
            self._rows = OrderedDict()

    @property
    def tracking_rows(self) -> bool:
        """ Public getter.  Returns True/False to denote whether or not this table is tracking row records

        To be clear on the distinction: just about any table might have rows.  However,
        that doesn't mean that the mygration system should be syncing rows for that table.
        self.tracking_rows == True denotes that the mygration system thinks that we
        should probably be syncing rows for this table.

        :returns: Whether or not the mygration system is tracking rows on the table
        """
        return self._tracking_rows

    @property
    def columns(self) -> Dict[str, Column]:
        """ Public getter.  Returns an ordered dictionary of table columns """
        return self._columns

    @property
    def indexes(self) -> Dict[str, Index]:
        """ Public getter.  Returns an ordered dictionary of table indexes """
        return self._indexes

    @property
    def constraints(self) -> Dict[str, Constraint]:
        """ Public getter.  Returns an ordered dictionary of table constraints """
        return self._constraints

    @property
    def primary(self) -> Index:
        """ Public getter.  Returns the index object for the primary key """
        return self._primary

    @property
    def rows(self) -> Dict[int, List[Union[str, int]]]:
        """ Public getter.  Returns an ordered dictionary with row data by id """
        return None if self._rows is None else self._rows

    def add_rows(self, rows: Rows) -> Union[str, bool]:
        """ Adds rows to the table

        The rows object has some flexibility in terms of columns: it doesn't just
        assume that a value is provided for every column in the table.  Rather,
        there can be a list of columns and only those columns have values (which
        supports the equivalent of MySQL INSERT queries which only specify values
        for some columns).

        Rows with errors will not be processed

        :returns: An error string if an error is encountered, otherwise True/False
        """
        self._schema_errors = None
        self._schema_warnings = None
        if not isinstance(rows, Rows):
            raise ValueError(
                f"Only objects of class mygrations.core.definitions.rows can be added as rows to a table.  Instead I received an object of class '{rows.__class__.__name__}'"
            )

        if rows.parsing_errors:
            self._global_errors.extend(rows.parsing_errors)
        if rows.parsing_warnings:
            self._global_warnings.extend(rows.parsing_warnings)

        self._tracking_rows = True
        if self._rows is None:
            self._rows = OrderedDict()

        # the rows object may have a list of columns.  If not use our own list of columns
        # remember that self._columns is an OrderedDict so converting its keys to a list
        # actually preserves the order (which is a requirement for us)
        columns = rows.columns if rows.num_explicit_columns else list(self._columns.keys())
        if 'id' not in columns:
            self._global_errors.append(
                "A column named 'id' is required to manage rows in the table, but the id column is missing in the rows for table %s"
                % (self._name, )
            )
            return
        id_index = columns.index('id')

        for values in rows.raw_rows:
            # rows without explicit columns must be checked for matching columns
            if not rows.num_explicit_columns and len(values) != len(columns):
                self._global_errors.append(
                    'Insert values has wrong number of values for table %s and row %s' % (self._name, values)
                )
                continue

            # we need to know the id of this record
            row_id = str(values[id_index])
            if not row_id:
                self._global_errors.append(
                    'Row is missing a value for the id column for table %s and row %s' % (self._name, values)
                )
                continue

            if row_id in self._rows:
                self._global_errors.append('Duplicate row id found for table %s and row %s' % (self.name, values))
                continue

            if not row_id:
                self._global_errors.append('Invalid row id of %s found for table %s' % (row_id, self.name))
                continue

            self._rows[row_id] = OrderedDict(zip(columns, values))
            self._rows[row_id]['id'] = row_id

        return True

    def add_raw_row(self, row: Dict[str, Union[str, int]]) -> Union[str, bool]:
        """ Adds a row into the table as a dictionary instead of a row object

        A bit of repetition here.  This is similar to what happens inside the main
        loop of self.add_rows, but different enough that I'm not trying to combine
        them right this second.

        :returns: An error string if an error is encountered, otherwise True/False
        """
        self._schema_errors = None
        self._schema_warnings = None
        self._tracking_rows = True
        if self._rows is None:
            self._rows = OrderedDict()

        row_id = str(row.get('id'))
        if not row_id:
            raise ValueError("Cannot manage records without an 'id' column and value")
        if row_id in self._rows:
            return 'Duplicate row id found for table %s and row %s' % (self._name, row)

        # make sure we have a value for every column in the row, and build an OrderedDict
        converted_row = OrderedDict()
        for column in self._columns.keys():
            if column in row:
                converted_row[column] = row[column]
            else:
                converted_row[column] = self._columns[column].default
        converted_row['id'] = row_id

        self._rows[row_id] = converted_row

    def column_before(self, column_name: str) -> Union[str, bool]:
        """ Returns the name of the column that comes before a given row.

        Returns true if the column is at the beginning of the table

        :returns: The name of the column before the given column, or True if at the beginning
        """
        # this works because we used an OrderedDict
        columns = [col for col in self.columns.keys()]

        if not column_name in columns:
            raise ValueError(
                "Cannot return column before %s because %s does not exist in table %s" %
                (column_name, column_name, self.name)
            )

        index = columns.index(column_name)
        if index == 0:
            return True

        return columns[index - 1]

    def column_is_indexed(self, column: Union[str, Column]) -> bool:
        """ Returns True/False to denote whether or not the column has a useable index """
        if type(column) != str:
            column = column.name
        if column not in self._columns:
            return False

        # best way to do this is with a set.  We'll keep a record of all indexed columns
        # the column is indexed if an index has that column in the first position
        if self._indexed_columns is None:
            self._indexed_columns = set([index.columns[0] for index in self._indexes.values()])
        return column in self._indexed_columns

    def __str__(self) -> str:
        return str(self.create())

    def create(self, nice=False):
        """ Returns a create table operation that can create this table

        :param nice: Whether or not to return a nicely formatted CREATE TABLE command
        :returns: A create table operation
        """
        return CreateTable(self, nice)

    def nice(self) -> str:
        return str(self.create(True))

    def add_column(self, column: Column, position=False):
        """ Adds a column to the table

        The value of position matches self.position from mygrations.formats.mysql.mygration.operations.add_column

        :param column: The column to add
        :param position: Where to put the column
        """
        self._schema_errors = None
        self._schema_warnings = None
        if column.name in self._columns:
            raise ValueError("Cannot add column %s because %s already exists" % (column.name, column.name))

        # putting it at the end is easy
        if not position:
            self._columns[column.name] = column
            return None

        # beginning is also easy
        if position == True:
            self._columns[column.name] = column
            self._columns.move_to_end(column.name, last=False)
            return None

        # now it is tricky
        found = False
        new_columns = OrderedDict()
        for key, value in self._columns.items():
            new_columns[key] = value
            if key == position:
                new_columns[column.name] = column
                found = True

        if not found:
            raise ValueError(
                "Cannot add column %s after %s because %s does not exist" % (column.name, position, position)
            )

        self._columns = new_columns
        return None

    def remove_column(self, column: Union[str, Column]):
        """ Removes a column from the table """
        self._schema_errors = None
        self._schema_warnings = None
        column_name = column if type(column) == str else column.name
        if not column_name in self._columns:
            raise ValueError("Cannot remove column %s because column %s does not exist" % (column_name, column_name))
        self._columns.pop(column_name, None)

    def change_column(self, new_column: Column):
        """ Changes a column.  This does not currently support renaming columns """
        self._schema_errors = None
        self._schema_warnings = None
        if not new_column.name in self._columns:
            raise ValueError(
                "Cannot modify column %s because column %s does not exist" % (new_column.name, new_column.name)
            )
        self._columns[new_column.name] = new_column

    def add_index(self, index: Index):
        """ Adds an index to the table """
        self._schema_errors = None
        self._schema_warnings = None
        if index.name in self._indexes:
            raise ValueError("Cannot add index %s because index %s already exists" % (index.name, index.name))
        if index.index_type == 'PRIMARY':
            self._primary = index
        self._indexes[index.name] = index

        if self._indexed_columns is not None:
            self._indexed_columns.add(index.columns[0])

    def remove_index(self, index: Union[str, Index]):
        """ Removes an index from the table """
        self._schema_errors = None
        self._schema_warnings = None
        index_name = index if type(index) == str else index.name
        if index_name not in self._indexes:
            raise ValueError("Cannot remove index %s because index %s does not exist" % (index_name, index_name))

        indexed_column = self._indexes[index_name].columns[0]
        self._indexes.pop(index_name, None)
        if self._indexed_columns is not None:
            self._indexed_columns.discard(indexed_column)

    def change_index(self, new_index: Index):
        """ Changes an index.  This does not currently support renaming """
        self._schema_errors = None
        self._schema_warnings = None
        if not new_index.name in self._indexes:
            raise ValueError(
                "Cannot modify index %s because index %s does not exist" % (new_index.name, new_index.name)
            )

        if self._indexed_columns is not None:
            self._indexed_columns.discard(self._indexes[new_index.name].columns[0])
        self._indexes[new_index.name] = new_index

        if self._indexed_columns is not None:
            self._indexed_columns.add(new_index.columns[0])

    def add_constraint(self, constraint: Constraint):
        """ Adds a constraint to the table """
        self._schema_errors = None
        self._schema_warnings = None
        if constraint.name in self._constraints:
            raise ValueError(
                "Cannot add constraint %s because constraint %s already exists" % (constraint.name, constraint.name)
            )
        self._constraints[constraint.name] = constraint

    def remove_constraint(self, constraint: Union[str, Constraint]):
        """ Removes an constraint from the table """
        self._schema_errors = None
        self._schema_warnings = None
        if type(constraint) != str:
            constraint = constraint.name
        if constraint not in self._constraints:
            raise ValueError(
                "Cannot remove constraint %s because constraint %s does not exist" % (constraint, constraint)
            )
        self._constraints.pop(constraint, None)

    def change_constraint(self, new_constraint):
        """ Changes a constraint. This does not currently support renaming. """
        self._schema_errors = None
        self._schema_warnings = None
        if not new_constraint.name in self._constraints:
            raise ValueError(
                "Cannot modify constraint %s because constraint %s does not exist" %
                (new_constraint.name, new_constraint.name)
            )
        self._constraints[new_constraint.name] = new_constraint

    def _loose_equal(self, val1: Union[str, int], val2: Union[str, int]) -> bool:
        """ Performs a looser comparison, as values might have different types depending on whether they came out of a database or file

        Returns true if the two values are equal, even if one is a string and the other an int.
        """
        # if we don't have a type mistmatch then this is easy
        if type(val1) == type(val2):
            return val1 == val2

        # otherwise see if we can cheat and just convert to strings
        return str(val1) == str(val2)

    def to(self, comparison_table, split_operations=False):
        """ Compares two tables to eachother and returns a list of operations which can bring the structure of the second in line with the first

        In other words, this pseudo code will make table have the same structure as comparison_table

        for operation in table.to( comparison_table ):
            table.apply( operation )

        if split_operations is True then a dict of migration operations will be returned to separate
        foreign key operations from everything else.  The dict will have up to three keys: [ 'removed_fks', 'fks', 'kitchen_sink' ]

        'fks' contains the alter statement operation needed to add/change/remove foreign keys
        'kitchen_sink' contains everything else

        If that division of labor seems arbitrary, it isn't: it is separated out that
        way due to the needs of the overall algorithm.

        If split_operations is false then a single alter table operation will be returned that encompasses all changes

        :param comparison_table: A table to find differences with
        :param split_operations: Whether to combine all operations in one alter table or separate them
        :type comparison_table: mygrations.formats.mysql.definitions.table
        :type split_operations: bool
        :returns: A list of operations to apply to table
        :rtype: list[mygrations.formats.mysql.mygrations.operations.*] | dict
        """
        raise NotImplementedError()

    def to_rows(self, from_table=None):
        """ Compares two tables to eachother and returns a list of operations which can bring the rows of this table in line with the other

        It's important to note (and probably important to think through and change eventually)
        that this method has the opposite meaning of `mygrations.formats.mysql.definitions.table.to()`
        That method is called on the `from` table and operations on the (required) `to` table.
        This method is called on the `to` table and can (optionally) be passed in a `from` table.

        :param from_table: A table to find differences with (or None)
        :type from_table: mygrations.formats.mysql.definitions.table
        :returns: A list of operations to apply to table
        :rtype: list[mygrations.formats.mysql.mygrations.operations.*]
        """
        raise NotImplementedError()

    def _differences(self, from_list: Dict[str, Any], to_list: Dict[str, Any]):
        """
        Calculates the difference between two OrderedDicts.

        https://codereview.stackexchange.com/a/176303/140581

        :param from_list: OrderedDict
        :param to_list: OrderedDict
        :return: (added, removed, overlap)
        """

        return ([key for key in to_list if key not in from_list], [key for key in from_list if key not in to_list],
                [key for key in from_list if key in to_list])

    def apply_operation(self, operation):
        """
        Applies an operation to the table

        :param operation: The operation to apply
        :type operation: mygrations.formats.mysql.mygration.operations.*
        """
        operation.apply_to_table(self)



import unittest
from .option import Option
class TestOption(unittest.TestCase):
    def test_can_create(self):
        option = Option('my_option', 'my_value')

        self.assertEquals('my_option', option.name)
        self.assertEquals('my_value', option.value)

    def test_required_value(self):
        option = Option('my_option', '')

        self.assertEquals(['Missing value for option my_option'], option.schema_errors)



class RemoveTable:
    """ Generates an SQL command to drop a table """
    def __init__(self, table_name):
        if type(table_name) != str:
            self._table_name = table_name.name
        else:
            self._table_name = table_name

    @property
    def table_name(self) -> str:
        """ Returns the name of the table. """
        return self._table_name

    def __str__(self):
        return 'DROP TABLE `%s`;' % self._table_name



import unittest
from .remove_table import RemoveTable
class RemoveTableTest(unittest.TestCase):
    def test_as_string(self):
        self.assertEquals("DROP TABLE `test_table`;", str(RemoveTable('test_table')))



import unittest
from .enable_checks import EnableChecks
class EnableChecksTest(unittest.TestCase):
    def test_as_string(self):
        self.assertEquals("SET FOREIGN_KEY_CHECKS=1;", str(EnableChecks()))



import unittest
from .add_column import AddColumn
from ..definitions.columns import String
class AddColumnTest(unittest.TestCase):
    def test_as_string(self):
        string = String('name', 'VARCHAR', length=255, default='', null=False)
        add_column = AddColumn(string, 'age')
        self.assertEquals("ADD `name` VARCHAR(255) NOT NULL DEFAULT '' AFTER `age`", str(add_column))



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_text import TypeText
class TestTypeText(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = TypeText()
        returned = parser.parse("name text not null,")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('name', parser._name)
        self.assertEquals('text', parser._column_type)
        self.assertFalse(parser._null)
        self.assertTrue(parser.has_comma)
        self.assertEquals('', parser._character_set)
        self.assertEquals('', parser._collate)
        self.assertEquals(0, len(parser._parsing_errors))
        self.assertEquals(0, len(parser._parsing_warnings))

    def test_character_set(self):

        # parse typical insert values
        parser = TypeText()
        returned = parser.parse("name text character set 'blah' collate 'boo',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('name', parser._name)
        self.assertEquals('text', parser._column_type)
        self.assertTrue(parser._null)
        self.assertTrue(parser.has_comma)
        self.assertEquals('blah', parser._character_set)
        self.assertEquals('boo', parser._collate)
        self.assertEquals(0, len(parser._parsing_errors))
        self.assertEquals(0, len(parser._parsing_warnings))

    def test_strip_backticks(self):

        # parse typical insert values
        parser = TypeText()
        returned = parser.parse("`name` text not null")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertEquals('name', parser._name)



import unittest

from mygrations.formats.mysql.file_reader.database import Database as DatabaseReader
from mygrations.formats.mysql.file_reader.create_parser import CreateParser
from mygrations.formats.mysql.mygrations.mygration import Mygration
class AddConflictingForeignKeyTests(unittest.TestCase):
    def test_add_conflicting_separates_fks(self):
        """ 3 adds. """

        table1 = """CREATE TABLE `tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
`account_id` INT(10) UNSIGNED NOT NULL,
`repeating_task_id` INT(10) UNSIGNED NOT NULL,
`name` VARCHAR(255) NOT NULL DEFAULT '',
PRIMARY KEY (`id`),
KEY `account_id_tasks` (`account_id`),
KEY `repeating_task_id_tasks` (`repeating_task_id`),
CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
CONSTRAINT `repeating_task_id_tasks_fk` FOREIGN KEY (`repeating_task_id`) REFERENCES `repeating_tasks` (`id`) ON DELETE CASCADE ON UPDATE CASCADE) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        table2 = """CREATE TABLE `repeating_tasks` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
`account_id` INT(10) UNSIGNED NOT NULL,
`task_id` INT(10) UNSIGNED NOT NULL,
`name` VARCHAR(255) NOT NULL DEFAULT '',
PRIMARY KEY (`id`),
KEY `account_id_rts` (`account_id`),
KEY `task_id_rts` (`task_id`),
CONSTRAINT `account_id_tasks_fk` FOREIGN KEY (`account_id`) REFERENCES `accounts` (`id`) ON DELETE CASCADE ON UPDATE CASCADE,
CONSTRAINT `task_id_rts` FOREIGN KEY (`task_id`) REFERENCES `tasks` (`id`) ON DELETE CASCADE ON UPDATE CASCADE) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        table3 = """CREATE TABLE `accounts` (`id` INT(10) UNSIGNED NOT NULL AUTO_INCREMENT,
`name` VARCHAR(255) NOT NULL DEFAULT '',
PRIMARY KEY (`id`)) ENGINE=InnoDB DEFAULT CHARSET=utf8;"""
        tables = [table1, table2, table3]
        db = DatabaseReader(tables)
        mygrate = Mygration(db)

        self.assertEquals([
            "Duplicate foreign key: foreign key named 'account_id_tasks_fk' exists in tables 'tasks' and 'repeating_tasks'"
        ], mygrate.errors)
        self.assertEquals(None, mygrate.operations)



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_character import TypeCharacter
class TestTypeCharacter(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = TypeCharacter()
        returned = parser.parse("name varchar(255) not null default 'sup',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('name', parser._name)
        self.assertEquals('varchar', parser._column_type)
        self.assertEquals('255', parser._length)
        self.assertFalse(parser._null)
        self.assertEquals('sup', parser._default)
        self.assertTrue(parser.has_comma)
        self.assertTrue(parser._character_set is None)
        self.assertTrue(parser._collate is None)

    def test_character_set(self):

        # parse typical insert values
        parser = TypeCharacter()
        returned = parser.parse("name varchar(255) not null default 'sup' character set 'blah' collate 'boo',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('name', parser._name)
        self.assertEquals('varchar', parser._column_type)
        self.assertEquals('255', parser._length)
        self.assertFalse(parser._null)
        self.assertEquals('sup', parser._default)
        self.assertTrue(parser.has_comma)
        self.assertEquals('blah', parser._character_set)
        self.assertEquals('boo', parser._collate)

    def test_optional_default(self):

        # parse typical insert values
        parser = TypeCharacter()
        returned = parser.parse("name varchar(255)")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('name', parser._name)
        self.assertEquals('varchar', parser._column_type)
        self.assertEquals('255', parser._length)
        self.assertTrue(parser._null)
        self.assertEquals(None, parser._default)
        self.assertFalse(parser.has_comma)
        self.assertTrue(parser._character_set is None)
        self.assertTrue(parser._collate is None)

    def test_empty_default(self):

        # parse typical insert values
        parser = TypeCharacter()
        returned = parser.parse("name varchar(255) default ''")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('', parser._default)

    def test_strip_backticks(self):

        # parse typical insert values
        parser = TypeCharacter()
        returned = parser.parse("`name` varchar(255)")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('name', parser._name)



from ..definitions.index import Index
from ..definitions.table import Table
class RemoveIndex:
    """ Generates a partial SQL command to drop an index from a table """
    def __init__(self, index: Index):
        self._index = index

    def apply_to_table(self, table: Table):
        """ Removes the index from the table """
        table.remove_index(self._index)

    def __str__(self):
        if self._index.index_type == 'PRIMARY':
            return 'DROP PRIMARY KEY'
        return 'DROP KEY `%s`' % (self._index.name)



from .column import Column
from typing import List, Union
class Numeric(Column):
    _allowed_column_types = [
        'INTEGER',
        'INT',
        'SMALLINT',
        'TINYINT',
        'MEDIUMINT',
        'BIGINT',
        'DECIMAL',
        'NUMERIC',
        'FLOAT',
        'DOUBLE',
        'BIT',
    ]

    def __init__(
        self,
        name: str = '',
        column_type: str = '',
        length: Union[str, int] = None,
        null: bool = True,
        has_default: bool = False,
        default: Union[str, int] = None,
        unsigned: bool = None,
        character_set: str = None,
        collate: str = None,
        auto_increment: bool = False,
        enum_values: List[str] = None,
        parsing_errors: List[str] = None,
        parsing_warnings: List[str] = None,
    ):
        # it would be nice to just do `def __init__(**kwargs)` and then `super().__init__(**kwargs)`
        # but then we would lose our type hints.  :shrug:
        super().__init__(
            name=name,
            column_type=column_type,
            length=length,
            null=null,
            has_default=has_default,
            default=default,
            unsigned=unsigned,
            character_set=character_set,
            collate=collate,
            auto_increment=auto_increment,
            enum_values=enum_values,
            parsing_errors=parsing_errors,
            parsing_warnings=parsing_warnings,
        )

    def _check_for_schema_errors_and_warnings(self):
        super()._check_for_schema_errors_and_warnings()

        allow_float = ['DECIMAL', 'NUMERIC', 'FLOAT', 'DOUBLE']
        no_length = ['FLOAT', 'DOUBLE', 'BIT']
        no_auto_increment = ['FLOAT', 'DOUBLE', 'BIT']

        # I should probably have the parser auto-convert the type based on quotes/whatever, but
        # currently it doesn't and I'm apparently being lazy.  I'll probably regret this.
        default_type = None
        if type(self.default) == int:
            default_type = 'int'
        elif type(self.default) == float:
            default_type = 'float'
        elif type(self.default) == str:
            default_type = 'str'
            try:
                int(self.default)
                default_type = 'int'
            except:
                try:
                    float(self.default)
                    default_type = 'float'
                except:
                    pass

        if default_type == 'str':
            self._schema_errors.append(
                f"Column '{self.name}' of type '{self.column_type}' cannot have a string value as a default"
            )
        else:
            if default_type == 'float' and self.column_type not in allow_float:
                self._schema_errors.append(
                    f"Column '{self.name}' of type '{self.column_type}' must have an integer value as a default"
                )
            if self.column_type == 'BIT' and int(self.default) != 0 and int(self.default) != 1:
                self._schema_errors.append(f"Column '{self.name}' of type 'BIT' must have a default of 1 or 0")

        if self.length:
            if self.column_type in no_length:
                self._schema_errors.append(f"Column '{self.name}' of type '{self.column_type}' cannot have a length")
            elif type(self.length) == str and ',' in self.length and self.column_type not in allow_float:
                self._schema_errors.append(
                    f"Column '{self.name}' of type '{self.column_type}' must have an integer value as its length"
                )

        if self.character_set is not None:
            self._schema_errors.append(f"Column '{self.name}' of type '{self.column_type}' cannot have a character set")
        if self.collate is not None:
            self._schema_errors.append(f"Column '{self.name}' of type '{self.column_type}' cannot have a collate")

        if self.auto_increment and self.column_type in no_auto_increment:
            self._schema_errors.append(f"Column '{self.name}' of type '{self.column_type}' cannot be an AUTO_INCREMENT")

        if self.enum_values:
            self._schema_errors.append(
                "Column '%s' of type %s is not allowed to have a list of values for its length" %
                (self.name, self.column_type)
            )

    def _is_really_the_same_default(self, column: Column) -> bool:
        if self.column_type != 'DECIMAL':
            return float(self.default) == float(column.default)

        # Default equality is mildly tricky for decimals because 0 and 0.000 are the same,
        # and if there are 4 digits after the decimal than 0.0000 and 0.00001 are the same too
        # This will come up if someone sets a default in an SQL file with too many (or too few) decimals,
        # while MySQL will report it properly rounded to the exact number of decimal places
        split = self.length.split(',')
        if len(split) == 2:
            ndecimals = int(split[1])
            if round(float(self.default), ndecimals) == round(float(column.default), ndecimals):
                return True

        return self.default == column.default



from .column import Column
from typing import List, Union
class Date(Column):
    _allowed_column_types = [
        'DATE',
        'DATETIME',
        'TIMESTAMP',
    ]

    def __init__(
        self,
        name: str = '',
        column_type: str = '',
        length: Union[str, int] = None,
        null: bool = True,
        has_default: bool = False,
        default: Union[str, int] = None,
        unsigned: bool = None,
        character_set: str = None,
        collate: str = None,
        auto_increment: bool = False,
        enum_values: List[str] = None,
        parsing_errors: List[str] = None,
        parsing_warnings: List[str] = None,
    ):
        # it would be nice to just do `def __init__(**kwargs)` and then `super().__init__(**kwargs)`
        # but then we would lose our type hints.  :shrug:
        super().__init__(
            name=name,
            column_type=column_type,
            length=length,
            null=null,
            has_default=has_default,
            default=default,
            unsigned=unsigned,
            character_set=character_set,
            collate=collate,
            auto_increment=auto_increment,
            enum_values=enum_values,
            parsing_errors=parsing_errors,
            parsing_warnings=parsing_warnings,
        )

    def _check_for_schema_errors_and_warnings(self):
        super()._check_for_schema_errors_and_warnings()

        if self.length:
            self._schema_errors.append(f"Column '{self.name}' of type '{self.column_type}' cannot have a length")

        if self.character_set is not None:
            self._schema_errors.append(f"Column '{self.name}' of type '{self.column_type}' cannot have a character set")
        if self.collate is not None:
            self._schema_errors.append(f"Column '{self.name}' of type '{self.column_type}' cannot have a collate")

        if self.auto_increment and self.column_type in no_auto_increment:
            self._schema_errors.append(f"Column '{self.name}' of type '{self.column_type}' cannot be an AUTO_INCREMENT")

        if self.values:
            self._schema_errors.append(
                "Column '%s' of type %s is not allowed to have a list of values for its length" %
                (self.name, self.column_type)
            )

        if self.unsigned:
            self._schema_errors.append("Column %s cannot be unsigned" % self._name)



from .column import Column
from typing import List, Union
class String(Column):
    _allowed_column_types = [
        'CHAR',
        'VARCHAR',
        'BINARY',
        'VARBINARY',
        'TINYBLOB',
        'BLOB',
        'MEDIUMBLOB',
        'LONGBLOB',
        'TINYTEXT',
        'TEXT',
        'MEDIUMTEXT',
        'LONGTEXT',
        'JSON',
    ]

    _allowed_default = [
        'CHAR',
        'VARCHAR',
        'BINARY',
        'VARBINARY',
        'ENUM',
        'SET',
    ]

    _allowed_collation = [
        'CHAR',
        'VARCHAR',
        'BINARY',
        'VARBINARY',
        'ENUM',
        'SET',
    ]

    _allowed_length = [
        'CHAR',
        'VARCHAR',
        'BINARY',
        'VARBINARY',
    ]

    _allowed_default_list = [
        'ENUM',
        'SET',
    ]

    def __init__(
        self,
        name: str = '',
        column_type: str = '',
        length: Union[str, int] = None,
        null: bool = True,
        has_default: bool = False,
        default: Union[str, int] = None,
        unsigned: bool = None,
        character_set: str = None,
        collate: str = None,
        auto_increment: bool = False,
        enum_values: List[str] = None,
        parsing_errors: List[str] = None,
        parsing_warnings: List[str] = None,
    ):
        # it would be nice to just do `def __init__(**kwargs)` and then `super().__init__(**kwargs)`
        # but then we would lose our type hints.  :shrug:
        super().__init__(
            name=name,
            column_type=column_type,
            length=length,
            null=null,
            has_default=has_default,
            default=default,
            unsigned=unsigned,
            character_set=character_set,
            collate=collate,
            auto_increment=auto_increment,
            enum_values=enum_values,
            parsing_errors=parsing_errors,
            parsing_warnings=parsing_warnings,
        )

    def __str__(self) -> str:
        as_string = super(String, self).__str__()
        if self.character_set:
            as_string += f" CHARACTER SET '{self.character_set}'"

        if self.collate:
            as_string += f" COLLATE '{self.collate}'"

        return as_string

    def _check_for_schema_errors_and_warnings(self):
        super()._check_for_schema_errors_and_warnings()

        if self.has_default and self.column_type not in self._allowed_default:
            self._schema_errors.append(f"Column '{self.name}' of type '{self.column_type}' cannot have a default")

        if self.auto_increment:
            self._schema_errors.append(
                f"Column '{self.name}' of type '{self.column_type}' cannot be an AUTO_INCREMENT: only numeric columns can"
            )

        if (self.character_set or self.collate):
            if self.column_type not in self._allowed_collation:
                self._schema_errors.append(
                    f"Column '{self.name}' of type '{self.column_type}' cannot have a collation/character set"
                )
            is_binary = self.column_type == 'BINARY' or self.column_type == 'VARBINARY'
            binary_collation = 'BINARY' if self.collate is None else self.collate
            binary_character_set = 'BINARY' if self.character_set is None else self.character_set
            if is_binary and (binary_collation != 'BINARY' or binary_character_set != 'BINARY'):
                self._schema_errors.append(
                    f"Column '{self.name}' of type '{self.column_type}' can only have a collate/character set of BINARY"
                )

        if self.length and self.column_type not in self._allowed_length:
            self._schema_errors.append(f'Column {self.name} of type {self.column_type} cannot have a length')

        if self.enum_values and self.column_type not in self._allowed_default_list:
            self._schema_errors.append(
                "Column '%s' of type %s is not allowed to have a list of values for its length" %
                (self.name, self.column_type)
            )

        if self.unsigned:
            self._schema_errors.append("Column %s cannot be unsigned" % self._name)

    def is_really_the_same_as(self, column: Column) -> bool:
        if not super().is_really_the_same_as(column):
            return False

        # if collate or character_set are different and *both* have a value,
        # then these aren't really the same
        for attr in ['collate', 'character_set']:
            my_val = getattr(self, attr)
            that_val = getattr(column, attr)
            if my_val and that_val and my_val != that_val:
                return False

        return True



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Union, List, Tuple
from .base import Base
class Constraint(Base):
    _column_name: str = ''
    _errors: List[str] = None
    _foreign_table_name: str = ''
    _foreign_column_name: str = ''
    _name: str = ''
    _on_delete: str = ''
    _on_update: str = ''
    _warnings: List[str] = None

    def __init__(
        self,
        name: str = '',
        column_name: str = '',
        foreign_table_name: str = '',
        foreign_column_name: str = '',
        on_delete: str = '',
        on_update: str = ''
    ):
        self._column_name = column_name
        self._foreign_table_name = foreign_table_name
        self._foreign_column_name = foreign_column_name
        self._name = name
        self._on_delete = on_delete.upper()
        self._on_update = on_update.upper()
        if not self._on_delete:
            self._on_delete = 'RESTRICT'
        if not self._on_update:
            self._on_update = 'RESTRICT'
        self._schema_errors = None
        self._schema_warnings = None
        self._parsing_errors = None
        self._parsing_warnings = None

    @property
    def name(self) -> str:
        """ Public getter.  Returns the name of the constraint. """
        return self._name

    @property
    def column_name(self) -> str:
        """ Public getter.  Returns the name of the column this constraint is on """
        return self._column_name

    @property
    def foreign_table_name(self) -> str:
        """ Public getter.  Returns the name of the table this constraint is to """
        return self._foreign_table_name

    @property
    def foreign_column_name(self) -> str:
        """ Public getter.  Returns the name of the column in the foreign table this constraint is to """
        return self._foreign_column_name

    @property
    def on_delete(self) -> str:
        """ Public getter.  Returns the ON DELETE action for this constraint """
        return self._on_delete

    @property
    def on_update(self) -> str:
        """ Public getter.  Returns the ON UPDATE action for this constraint """
        return self._on_update

    @property
    def schema_errors(self) -> List[str]:
        """ Returns a list of schema errors """
        if self._schema_errors is None:
            self._check_for_schema_errors_and_warnings()
        return self._schema_errors

    @property
    def schema_warnings(self) -> List[str]:
        """ Returns a list of schema warnings """
        if self._schema_errors is None:
            self._check_for_schema_errors_and_warnings()
        return self._schema_warnings

    @property
    def parsing_errors(self) -> List[str]:
        """ Returns a list of parsing errors """
        return self._parsing_errors if self._parsing_errors is not None else []

    @property
    def parsing_warnings(self) -> List[str]:
        """ Returns a list of parsing/table warnings """
        return self._parsing_warnings if self._parsing_warnings is not None else []

    def _check_for_schema_errors_and_warnings(self) -> List[str]:
        self._schema_errors = []
        self._schema_warnings = []

        allowed_actions = ['CASCADE', 'NO ACTION', 'RESTRICT', 'SET DEFAULT', 'SET NULL']
        if not self.name:
            self._schema_errors.append('Missing name for constraint')
        elif len(self._name) > 64:
            self._schema_errors.append(f"Key name '{self.name}' is too long")

        if self.on_delete not in allowed_actions:
            self._schema_errors.append(
                f"ON DELETE action of '{self.on_delete}' for constraint '{self.name}' is not a valid ON DELETE action"
            )

        if self.on_update not in allowed_actions:
            self._schema_errors.append(
                f"ON UPDATE action of '{self.on_update}' for constraint '{self.name}' is not a valid ON UPDATE action"
            )

        for required in ['name', 'column_name', 'foreign_table_name', 'foreign_column_name']:
            if not getattr(self, required):
                self._schema_errors.append(f"Missing {required} for constraint '{self.name}'")

    def __str__(self) -> str:
        """ Returns the MySQL command that would create the constraint

        i.e. CONSTRAINT `vendors_w9_fk` FOREIGN KEY (`w9_id`) REFERENCES `vendor_w9s` (`id`) ON UPDATE CASCADE

        :returns: A partial MySQL command that could be used to generate the foreign key
        :rtype: string
        """
        return ' '.join([
            'CONSTRAINT',
            f'`{self.name}`',
            'FOREIGN KEY',
            f'(`{self.column_name}`)',
            'REFERENCES',
            f'`{self.foreign_table_name}`',
            f'(`{self.foreign_column_name}`)',
            f'ON DELETE {self.on_delete}',
            f'ON UPDATE {self.on_update}',
        ])



from typing import Union
from ..definitions.columns.column import Column
from ..definitions.table import Table
class AddColumn:
    _column: Column
    _position: Union[str, bool] = None
    """ Generates a partial SQL command to add a column to a table """
    def __init__(self, column: Column, position: Union[str, bool] = None):
        """ Create an add column operation

        Position should be one of:

        ==================  ====================
        Value               Meaning
        ==================  ====================
        None                Put at the end of the table
        False               Put at the end of the table
        True                Put at the beginning of the table
        Any string          position should be a column name: put the new column after that one
        ==================  ====================

        :param column: The column to add
        :param position: Where to put the new column in the table
        :type column: mygrations.formats.mysql.definitions.column
        :type position: mixed:
        """
        self._column = column
        self._position = position

    def apply_to_table(self, table: Table):
        """ Adds the column to the table """
        table.add_column(self._column, self._position)

    def __str__(self):
        if self._position == True:
            position = ' FIRST'
        elif type(self._position) == str:
            position = ' AFTER `%s`' % self._position
        else:
            position = ''

        return 'ADD %s%s' % (str(self._column), position)



from mygrations.core.parse.parser import Parser
from .type import Type
class TypeNumeric(Parser, Type):

    allowed_types = {
        'bit': True,
        'tinyint': True,
        'smallint': True,
        'mediumint': True,
        'int': True,
        'integer': True,
        'bigint': True,
        'decimal': True,
        'numeric': True,
        'char': True,
        'varchar': True
    }

    has_comma = False
    is_char = False

    # `created` int(10) unsigned not null default 0 AUTO_INCREMENT
    rules = [{
        'type': 'regexp',
        'value': '[^\(\s\)]+',
        'name': 'name'
    }, {
        'type': 'regexp',
        'value': '\w+',
        'name': 'type'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type': 'regexp',
        'value': '\d+',
        'name': 'length'
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'literal',
        'value': 'UNSIGNED',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'NOT NULL',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'DEFAULT ([^\(\s\),]+)',
        'name': 'default',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'AUTO_INCREMENT',
        'optional': True
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def process(self):

        self.has_comma = True if 'ending_comma' in self._values else False

        self._parsing_errors = []
        self._parsing_warnings = []
        self._schema_errors = []
        self._schema_warnings = []
        self._name = self._values['name'].strip('`')
        self._column_type = self._values['type']
        self._length = self._values['length']
        self._unsigned = True if 'UNSIGNED' in self._values else False
        self._null = False if 'NOT NULL' in self._values else True
        self._has_default = 'default' in self._values
        self._default = self._values['default'] if 'default' in self._values else None
        self._auto_increment = True if 'AUTO_INCREMENT' in self._values else False
        self.is_char = self._column_type in ['char', 'varchar']

        # make sense of the default
        if self._default and len(self._default) >= 2 and self._default[0] == "'" and self._default[-1] == "'":
            self._default = self._default.strip("'")
        elif self._default:
            if self._default.lower() == 'null':
                self._default = None
            elif self.is_char:
                self._parsing_warnings.append('Default value for column %s should be quoted' % self._name)



from collections import OrderedDict
from .rows import Rows as RowsDefinition
from ..mygrations.operations.alter_table import AlterTable
from ..mygrations.operations.add_column import AddColumn
from ..mygrations.operations.change_column import ChangeColumn
from ..mygrations.operations.remove_column import RemoveColumn
from ..mygrations.operations.add_key import AddKey
from ..mygrations.operations.change_key import ChangeKey
from ..mygrations.operations.remove_key import RemoveKey
from ..mygrations.operations.add_constraint import AddConstraint
from ..mygrations.operations.change_constraint import ChangeConstraint
from ..mygrations.operations.remove_constraint import RemoveConstraint
from ..mygrations.operations.create_table import CreateTable
from ..mygrations.operations.row_delete import RowDelete
from ..mygrations.operations.row_insert import RowInsert
from ..mygrations.operations.row_update import RowUpdate
from mygrations.core.definitions.table import Table as BaseTable
class Table(BaseTable):
    def create(self, nice=False):
        """ Returns a create table operation that can create this table

        :param nice: Whether or not to return a nicely formatted CREATE TABLE command
        :type nice: bool
        :returns: A create table operation
        :rtype: mygrations.operations.CreateTable
        """
        return CreateTable(self, nice)

    def to(self, comparison_table, split_operations=False):
        """ Compares two tables to eachother and returns a list of operations which can bring the structure of the second in line with the first

        In other words, this pseudo code will make table have the same structure as comparison_table

        for operation in table.to( comparison_table ):
            table.apply( operation )

        if split_operations is True then a dict of migration operations will be returned to separate
        foreign key operations from everything else.  The dict will have up to three keys: [ 'removed_fks', 'fks', 'kitchen_sink' ]

        'fks' contains the alter statement operation needed to add/change/remove foreign keys
        'kitchen_sink' contains everything else

        If that division of labor seems arbitrary, it isn't: it is separated out that
        way due to the needs of the overall algorithm.

        If split_operations is false then a single alter table operation will be returned that encompasses all changes

        :param comparison_table: A table to find differences with
        :param split_operations: Whether to combine all operations in one alter table or separate them
        :type comparison_table: mygrations.formats.mysql.definitions.table
        :type split_operations: bool
        :returns: A list of operations to apply to table
        :rtype: list[mygrations.formats.mysql.mygrations.operations.*] | dict
        """
        # start with the columns, obviously
        (added_columns, removed_columns, overlap_columns) = self._differences(self.columns, comparison_table.columns)
        (added_keys, removed_keys, overlap_keys) = self._differences(self.indexes, comparison_table.indexes)
        (added_constraints, removed_constraints,
         overlap_constraints) = self._differences(self.constraints, comparison_table.constraints)

        # keeping in mind the overall algorithm, we're going to separate out all changes into three alter statments
        # these are broken up according to the way that the system has to process them to make sure that foreign
        # keys are not violated during the process
        # 1. Adding columns, changing columns, adding keys, changing keys, removing keys, removing foreign keys
        # 2. Adding foreign keys, changing foreign keys
        # 3. Removing columns
        primary_alter = AlterTable(self.name)
        for new_column in added_columns:
            primary_alter.add_operation(
                AddColumn(comparison_table.columns[new_column], comparison_table.column_before(new_column))
            )

        for overlap_column in overlap_columns:
            # it's really easy to tell if a column changed
            if str(self.columns[overlap_column]) == str(comparison_table.columns[overlap_column]):
                continue

            # FALSE POSITIVE CHECK:
            # if one column has collate/character set and the other doesn't, and that is the only difference,
            # then ignore this difference
            if self.columns[overlap_column].is_really_the_same_as(comparison_table.columns[overlap_column]):
                continue

            primary_alter.add_operation(ChangeColumn(comparison_table.columns[overlap_column]))

        for removed_column in removed_columns:
            primary_alter.add_operation(RemoveColumn(self.columns[removed_column]))

        # indexes also go in that first alter table
        for new_key in added_keys:
            primary_alter.add_operation(AddKey(comparison_table.indexes[new_key]))
        for removed_key in removed_keys:
            primary_alter.add_operation(RemoveKey(self.indexes[removed_key]))
        for overlap_key in overlap_keys:
            if str(self.indexes[overlap_key]) == str(comparison_table.indexes[overlap_key]):
                continue
            primary_alter.add_operation(ChangeKey(comparison_table.indexes[overlap_key]))

        # removed foreign key constraints get their own alter because that should always happen first
        removed_constraints_alter = AlterTable(self.name)
        for removed_constraint in removed_constraints:
            removed_constraints_alter.add_operation(RemoveConstraint(self.constraints[removed_constraint]))

        # adding/changing/removing foreign keys gets their own alter
        constraints = AlterTable(self.name)
        for added_constraint in added_constraints:
            constraints.add_operation(AddConstraint(comparison_table.constraints[added_constraint]))
        for overlap_constraint in overlap_constraints:
            if str(self.constraints[overlap_constraint]) == str(comparison_table.constraints[overlap_constraint]):
                continue

            # foreign key constraints are modified by first dropping the constraint and
            # then adding the new one.  However, these two operations cannot happen in the
            # same alter command.  Kinda a pain.  Oh well.
            removed_constraints_alter.add_operation(RemoveConstraint(self.constraints[overlap_constraint]))
            constraints.add_operation(AddConstraint(comparison_table.constraints[overlap_constraint]))

        # now put it all together
        if split_operations:
            operations = {}
            if removed_constraints_alter:
                operations['removed_fks'] = removed_constraints_alter
            if primary_alter:
                operations['kitchen_sink'] = primary_alter
            if constraints:
                operations['fks'] = constraints
        else:
            operations = []
            for operation in constraints:
                primary_alter.add_operation(operation)
            if removed_constraints_alter:
                operations.append(removed_constraints_alter)
            if primary_alter:
                operations.append(primary_alter)

        return operations

    def to_rows(self, from_table=None):
        """ Compares two tables to eachother and returns a list of operations which can bring the rows of this table in line with the other

        It's important to note (and probably important to think through and change eventually)
        that this method has the opposite meaning of `mygrations.formats.mysql.definitions.table.to()`
        That method is called on the `from` table and operations on the (required) `to` table.
        This method is called on the `to` table and can (optionally) be passed in a `from` table.

        :param from_table: A table to find differences with (or None)
        :type from_table: mygrations.formats.mysql.definitions.table
        :returns: A list of operations to apply to table
        :rtype: list[mygrations.formats.mysql.mygrations.operations.*]
        """
        if from_table and not from_table.tracking_rows:
            raise ValueError(
                "Refusing to compare rows for table %s that is not tracking rows.  Technically I can, but this is probably a sign that you are doing something wrong"
                % (self.name)
            )

        (inserted_ids, deleted_ids, updated_ids) = self._differences(from_table.rows if from_table else {}, self.rows)

        operations = []
        for row_id in inserted_ids:
            operations.append(RowInsert(self.name, self.rows[row_id]))

        for row_id in deleted_ids:
            operations.append(RowDelete(self.name, row_id))

        for row_id in updated_ids:
            # try to be smart and not update if we don't have to
            (inserted_cols, deleted_cols, updated_cols) = self._differences(self.rows[row_id], from_table.rows[row_id])
            differences = False
            for col in updated_cols:
                if not self._loose_equal(self.rows[row_id][col], from_table.rows[row_id][col]):
                    differences = True
                    break

            if differences:
                operations.append(RowUpdate(self.name, self.rows[row_id]))

        return operations



from .constraint_foreign import ConstraintForeign
from .index_key import IndexKey
from .index_primary import IndexPrimary
from .index_unique import IndexUnique
from .type_character import TypeCharacter
from .type_datetime import TypeDateTime
from .type_decimal import TypeDecimal
from .type_enum import TypeEnum
from .type_numeric import TypeNumeric
from .type_plain import TypePlain
from .type_text import TypeText
from .table_option import TableOption
from .insert_values import InsertValues

__all__ = [
    "ConstraintForeign", "IndexKey", "IndexPrimary", "IndexUnique", "TypeCharacter", "TypeDecimal", "TypeDateTime",
    "TypeEnum", "TypeNumeric", "TypePlain", "TypeText", "TableOption", "InsertValues"
]



import unittest
from .add_row import AddRow
from collections import OrderedDict
class AddRowTest(unittest.TestCase):
    def test_as_string(self):
        row = OrderedDict([('id', 10), ('age', '5'), ('dob', None)])
        add_row = AddRow('test_table', row)
        self.assertEquals("INSERT INTO `test_table` (`id`, `age`, `dob`) VALUES ('10', '5', NULL);", str(add_row))



import unittest
from .change_constraint import ChangeConstraint
from ..definitions.constraint import Constraint
class ChangeConstraintTest(unittest.TestCase):
    def test_as_string(self):
        constraint = Constraint('user_id_fk', 'user_id', 'users', 'id', on_update='cascade')
        self.assertEquals(
            "DROP FOREIGN KEY `user_id_fk`, ADD CONSTRAINT `user_id_fk` FOREIGN KEY (`user_id`) REFERENCES `users` (`id`) ON DELETE RESTRICT ON UPDATE CASCADE",
            str(ChangeConstraint(constraint))
        )



import unittest
from .column import Column
class MockColumn(Column):
    _allowed_column_types = ['INT']
class TestColumn(unittest.TestCase):
    def test_string_conversion(self):
        column = MockColumn(
            'test_column',
            'INT',
            length=25,
            null=False,
            default=0,
            unsigned=True,
        )

        self.assertEquals('`test_column` INT(25) UNSIGNED NOT NULL DEFAULT 0', str(column))

    def test_is_the_same(self):
        attrs = {
            'length': 5,
            'null': False,
            'unsigned': True,
            'default': '5',
        }
        column1 = MockColumn('test_column', 'INT', **attrs)
        column2 = MockColumn('test_column', 'INT', **attrs)

        self.assertTrue(column1.is_really_the_same_as(column2))
        self.assertTrue(column2.is_really_the_same_as(column1))

        diff = {
            'length': 6,
            'null': None,
            'unsigned': False,
            'default': 5,
        }
        for (key, val) in diff.items():
            column1 = MockColumn('test_column', 'INT', **attrs)
            column2 = MockColumn('test_column', 'INT', **{**attrs, **{key: val}})
            self.assertFalse(column1.is_really_the_same_as(column2))
            self.assertFalse(column2.is_really_the_same_as(column1))

        column1 = MockColumn('test_column', 'INT', **attrs)
        column2 = MockColumn('test_column', 'INT', **{**attrs, 'default': None})
        self.assertFalse(column1.is_really_the_same_as(column2))
        self.assertFalse(column2.is_really_the_same_as(column1))

    def test_not_null_needs_default(self):
        column = MockColumn(
            'test_column',
            'INT',
            length=25,
            null=False,
            default=None,
        )
        self.assertEquals(
            [
                'Column test_column does not allow null values and has no default: you should set a default to avoid warnings'
            ],
            column.schema_warnings,
        )

        column = MockColumn('test_column', 'INT')
        self.assertEquals([], column.schema_warnings)



import unittest
from .change_index import ChangeIndex
from ..definitions.index import Index
class ChangeIndexTest(unittest.TestCase):
    def test_as_string(self):
        index = Index('user_id', ['user_id', 'account_id'], 'index')
        self.assertEquals("DROP KEY `user_id`, ADD KEY `user_id` (`user_id`,`account_id`)", str(ChangeIndex(index)))



from __future__ import annotations
from typing import Union, List
from .columns.column import Column
class Rows:
    _columns: List[Column] = None
    _schema_errors: List[str] = None
    _schema_warnings: List[str] = None
    _parsing_errors: List[str] = None
    _parsing_warnings: List[str] = None
    _num_explicit_columns: int = 0
    _raw_rows: List[List[str]] = ''
    _table: str = ''
    _warnings: List[str] = None

    def __init__(
        self, table: str = '', raw_rows: List[List[str]] = [], columns: List[Column] = None, num_explicit_columns=0
    ):
        self._columns = columns if columns is not None else []
        self._schema_errors = []
        self._schema_warnings = []
        self._parsing_errors = []
        self._parsing_warnings = []
        self._num_explicit_columns = num_explicit_columns
        self._raw_rows = raw_rows
        self._table = table
        self._warnings = []

    @property
    def table(self) -> str:
        """ Returns the name of the table that records are being inserted for

        :returns: The name of the table
        """
        return self._table

    @property
    def raw_rows(self) -> List[List[str]]:
        """ Returns a list of insert values from the VALUES part of the query

        :returns: A list of values for each row
        """
        return self._raw_rows

    @property
    def columns(self) -> List[Column]:
        """ Returns the list of columns for the rows

        :returns: A list of columns
        """
        return self._columns

    @property
    def num_explicit_columns(self) -> int:
        """ Returns the number of columns specified in the insert query

        Can be zero if none are specified, which happens for queries like:
        INSERT INTO table (val1, val2, val3...);

        :returns: The number of columns which have been explicitly defined
        """
        return self._num_explicit_columns

    @property
    def schema_errors(self) -> List[str]:
        """ Returns a list of schema errors """
        return [] if self._schema_errors is None else self._schema_errors

    @property
    def schema_warnings(self) -> List[str]:
        """ Returns a list of schema warnings """
        return [] if self._schema_warnings is None else self._schema_warnings

    @property
    def parsing_errors(self) -> List[str]:
        """ Returns a list of parsing errors """
        return [] if self._parsing_errors is None else self._parsing_errors

    @property
    def parsing_warnings(self) -> List[str]:
        """ Returns a list of parsing warnings """
        return [] if self._parsing_warnings is None else self._parsing_warnings



from __future__ import annotations
from typing import Union, List
from ..base import Base
class Column(Base):
    _auto_increment: bool = None
    _character_set: str = None
    _collate: str = None
    _column_type: str = None
    _has_default: bool = False
    _default: Union[str, int] = None
    _length: Union[str, int] = None
    _name: str = ''
    _null: bool = None
    _unsigned: bool = None
    _allowed_column_types = []
    _enum_values: List[str] = None    # enum/set only
    _parsing_errors: List[str] = None
    _parsing_warnings: List[str] = None

    def __init__(
        self,
        name: str = '',
        column_type: str = '',
        length: Union[str, int] = None,
        null: bool = True,
        has_default: bool = False,
        default: Union[str, int] = None,
        unsigned: bool = None,
        character_set: str = None,
        collate: str = None,
        auto_increment: bool = False,
        enum_values: List[str] = None,
        parsing_errors: List[str] = None,
        parsing_warnings: List[str] = None,
    ):
        self._auto_increment = auto_increment
        self._character_set = character_set
        self._collate = collate
        self._column_type = column_type
        self._has_default = has_default
        self._default = default
        self._length = length
        self._name = name
        self._null = null
        self._unsigned = unsigned
        self._enum_values = enum_values
        self._parsing_errors = parsing_errors
        self._parsing_warnings = parsing_warnings
        self._schema_errors = None
        self._schema_warnings = None

    @property
    def name(self) -> str:
        """ Returns the name of the column. """
        return self._name

    @property
    def length(self) -> Union[str, int]:
        """ Returns the length of the column as a string or int.

        Some examples of the length for various column definitions:

        ==================  ====================
        Type                Length
        ==================  ====================
        INT(10) UNSIGNED    10
        VARCHAR(255)        255
        decimal(20,5)       20,5
        date
        ==================  ====================

        :returns: The column length
        """

        return self._length

    @property
    def null(self) -> bool:
        """ Returns True/False to denote if the column is allowed to be null """
        return self._null

    @property
    def column_type(self) -> str:
        """ Returns a string denoting the type of the column.  Always returns in uppercase

        Some examples of the length for various column definitions:

        ==================  ====================
        Definition          Type
        ==================  ====================
        INT(10) UNSIGNED    INT
        VARCHAR(255)        VARCHAR
        decimal(20,5)       DECIMAL
        date                DATE
        ==================  ====================
        """
        return self._column_type.upper()

    @property
    def has_default(self) -> bool:
        """ Returns True/False to denote if the column has a default value defined """
        return self._has_default

    @property
    def default(self) -> Union[str, int]:
        """ Returns the default value for the column as a string, or None for a default value of null

        Returns None to represent a default value of null.
        """
        return self._default

    @property
    def unsigned(self) -> bool:
        """ Returns True, False, or None to denote the status of the UNSIGNED property

        ==================  ====================
        Return Value        Meaning
        ==================  ====================
        True                The column is unsigned
        False               The column is signed
        None                UNSIGNED is not an applicable property for this column type
        ==================  ====================
        """
        return self._unsigned

    @property
    def character_set(self) -> str:
        """ Returns None or a value to denote the CHARACTER_SET property

        :returns: string, or None
        """
        return None if self._character_set is None else self._character_set.upper()

    @property
    def collate(self) -> str:
        """ Returns None or a value to denote the COLLATE property

        :returns: string, or None
        """
        return None if self._collate is None else self._collate.upper()

    @property
    def auto_increment(self) -> bool:
        """ Returns True, False, or None to denote the status of the AUTO_INCREMENT property

        ==================  ====================
        Return Value        Meaning
        ==================  ====================
        True                The column is an AUTO_INCREMENT column
        False               The column is not an AUTO_INCREMENT column
        None                AUTO_INCREMENT is not an applicable property for this column type
        ==================  ====================
        """
        return self._auto_increment

    @property
    def enum_values(self) -> List[str]:
        """ Returns the allowed values for the column (enum/set only)"""
        return self._enum_values

    @property
    def schema_errors(self) -> List[str]:
        """ Returns a list of schema errors """
        if self._schema_errors is None:
            self._check_for_schema_errors_and_warnings()
        return self._schema_errors

    @property
    def schema_warnings(self) -> List[str]:
        """ Returns a list of schema warnings """
        if self._schema_errors is None:
            self._check_for_schema_errors_and_warnings()
        return self._schema_warnings

    @property
    def parsing_errors(self) -> List[str]:
        """ Returns a list of parsing errors """
        return self._parsing_errors if self._parsing_errors is not None else []

    @property
    def parsing_warnings(self) -> List[str]:
        """ Returns a list of parsing warnings """
        return self._parsing_warnings if self._parsing_warnings is not None else []

    def _check_for_schema_errors_and_warnings(self):
        """ Finds and stores any schema errors/warnings in the appropriate parameters """
        self._schema_errors = []
        self._schema_warnings = []

        if not self.name:
            self._schema_errors.append("Column does not have a name")
        if not self.column_type:
            self._schema_errors.append(f"Missing column_type for column {self.name}")

        if self.default is None and not self.null and not self.auto_increment:
            self._schema_warnings.append(
                f'Column {self.name} does not allow null values and has no default: you should set a default to avoid warnings'
            )

    def __str__(self) -> str:
        """ Returns the MySQL command that would create the column

        i.e. column_name type(len) default ''

        :returns: A partial MySQL command that could be used to generate the column
        """
        parts = []
        parts.append('`%s`' % self.name)

        type_string = self.column_type
        if self.enum_values:
            type_string += "('%s')" % ("', '".join(self.enum_values))
        elif self.length is not None and self.length != '':
            type_string += '(%s)' % self.length
        parts.append(type_string)

        if self.unsigned:
            parts.append('UNSIGNED')

        if not self.null:
            parts.append('NOT NULL')

        if self.default is not None:
            if self.default == '':
                parts.append("DEFAULT ''")
            elif type(self.default) != str or self.default.isnumeric():
                parts.append("DEFAULT %s" % self.default)
            else:
                parts.append("DEFAULT '%s'" % self.default)

        if self.auto_increment:
            parts.append('AUTO_INCREMENT')

        return ' '.join(parts)

    def is_really_the_same_as(self, column: Column) -> bool:
        """ Takes care of a pesky false-positive when checking columns

        :param column: The column to comprehensively check for a difference with
        :returns: True if the column really is the same, even for apparent differences
        """
        # if any of these attributes change then it really isn't the same
        for attr in ['name', 'length', 'null', 'column_type', 'unsigned']:
            if getattr(self, attr) != getattr(column, attr):
                return False

        if self._default_none_mismatch(column):
            return False
        if not self._is_really_the_same_default(column):
            return False

        # if we got here then these columns are the same. Either all attributes have the same
        # values, or collate and/or character_set differ, but they differ by one not having
        # a value.  This is the false-positive we are trying to check for
        # (see tests.formats.mysql.definitions.test_table_text_false_positives)
        return True

    def _default_none_mismatch(self, column: Column) -> bool:
        if self.default is not None and column.default is None:
            return True
        if column.default is not None and self.default is None:
            return True
        return False

    def _is_really_the_same_default(self, column: Column) -> bool:
        return self.default == column.default



from ..definitions.constraint import Constraint
from ..definitions.table import Table
class AddConstraint:
    """ Generates a partial SQL command to add a FK to a table """
    _constraint: Constraint = None

    def __init__(self, constraint: Constraint):
        self._constraint = constraint

    def apply_to_table(self, table: Table):
        """ Adds the constraint to the table """
        table.add_constraint(self._constraint)

    def __str__(self):
        return 'ADD %s' % str(self._constraint)



from mygrations.core.parse.parser import Parser
from .type import Type
class TypeDecimal(Parser, Type):

    allowed_types = {'real': True, 'double': True, 'float': True, 'decimal': True, 'numeric': True}

    decimals = ''
    has_comma = False

    # longitude float(20,4) unsigned default null
    rules = [{
        'type': 'regexp',
        'value': '[^\(\s\)]+',
        'name': 'name'
    }, {
        'type': 'regexp',
        'value': '\w+',
        'name': 'type'
    }, {
        'type': 'literal',
        'value': '('
    }, {
        'type': 'regexp',
        'value': '\d+',
        'name': 'length'
    }, {
        'type': 'literal',
        'value': ',',
        'name': 'separating_comma'
    }, {
        'type': 'regexp',
        'value': '\d+',
        'name': 'decimals'
    }, {
        'type': 'literal',
        'value': ')'
    }, {
        'type': 'literal',
        'value': 'UNSIGNED',
        'optional': True
    }, {
        'type': 'literal',
        'value': 'NOT NULL',
        'optional': True
    }, {
        'type': 'regexp',
        'value': 'DEFAULT ([^\(\s\),]+)',
        'optional': True,
        'name': 'default'
    }, {
        'type': 'literal',
        'value': ',',
        'optional': True,
        'name': 'ending_comma'
    }]

    def process(self):

        self.has_comma = True if 'ending_comma' in self._values else False

        self._parsing_errors = []
        self._parsing_warnings = []
        self._schema_errors = []
        self._schema_warnings = []
        self._name = self._values['name'].strip('`')
        self._column_type = self._values['type']
        self.decimals = self._values['decimals']
        self._length = '%s,%s' % (self._values['length'], self._values['decimals'])
        self._unsigned = True if 'UNSIGNED' in self._values else False
        self._has_default = 'default' in self._values
        self._null = False if 'NOT NULL' in self._values else True
        self._default = self._values['default'] if 'default' in self._values else None

        # make sense of the default
        if self._default and len(self._default) >= 2 and self._default[0] == "'" and self._default[-1] == "'":
            self._default = self._default.strip("'")
            self._parsing_warnings.append(
                "Column '%s' has a numeric type but its default value is a string" % self._name
            )
        elif self._default and self._default.lower() == 'null':
            self._default = None



import unittest
from .disable_checks import DisableChecks
class DisableChecksTest(unittest.TestCase):
    def test_as_string(self):
        self.assertEquals("SET FOREIGN_KEY_CHECKS=0;", str(DisableChecks()))



import unittest
from .add_index import AddIndex
from ..definitions.index import Index
class AddIndexTest(unittest.TestCase):
    def test_as_string(self):
        index = Index('user_id', ['user_id', 'account_id'], 'index')
        add_index = AddIndex(index)
        self.assertEquals("ADD KEY `user_id` (`user_id`,`account_id`)", str(add_index))



import unittest
from .add_constraint import AddConstraint
from ..definitions.constraint import Constraint
class AddConstraintTest(unittest.TestCase):
    def test_as_string(self):
        constraint = Constraint('user_id_fk', 'user_id', 'users', 'id')
        add_constraint = AddConstraint(constraint)
        self.assertEquals(
            "ADD CONSTRAINT `user_id_fk` FOREIGN KEY (`user_id`) REFERENCES `users` (`id`) ON DELETE RESTRICT ON UPDATE RESTRICT",
            str(add_constraint)
        )



from typing import Union
from ..definitions.table import Table
class RemoveRow:
    """ Generates an SQL command to delete a record """
    def __init__(self, table_name: Union[str, Table], row_id: Union[int, str]):
        if type(table_name) != str:
            self._table_name = table_name.name
        else:
            self._table_name = table_name
        self.row_id = row_id

    @property
    def table_name(self) -> str:
        """ Returns the name of the table. """
        return self._table_name

    def __str__(self):
        return "DELETE FROM `%s` WHERE id='%s';" % (self._table_name, self.row_id)



from .plan import Plan
def execute(options, print_results=True):

    obj = Apply(options)
    return obj.execute(print_results=print_results)
class Apply(Plan):
    needs_cursor = False

    def execute(self, print_results):

        commands = self.build_commands()
        to_apply = ['%s' % command for command in commands]
        connection = self.get_driver()
        connection.execute(to_apply)
        if print_results:
            if to_apply:
                print('Applied:')
                print("\n".join(to_apply))
            else:
                print('No changes required')
        return (to_apply, True)



import unittest
from .index import Index
from .columns.string import String
class TestIndex(unittest.TestCase):
    def setUp(self):
        self.columns = ['name', 'description']

    def test_can_create(self):
        name = String('name', 'VARCHAR')
        description = String('description', 'VARCHAR')
        index = Index('test_index', self.columns, index_type='unique')

        self.assertEquals('test_index', index.name)
        self.assertEquals('UNIQUE', index.index_type)
        self.assertEquals(self.columns, index.columns)
        self.assertEquals([], index.schema_errors)
        self.assertEquals([], index.schema_warnings)
        self.assertEquals([], index.parsing_errors)
        self.assertEquals([], index.parsing_warnings)

    def test_string_conversion(self):
        index = Index(name='test_index', columns=self.columns, index_type='UNIQUE')

        self.assertEquals('UNIQUE KEY `test_index` (`name`,`description`)', str(index))

    def test_name_too_long(self):
        too_long = '012345678901234567890123456789012345678901234567890123456789012345'
        index = Index(too_long, self.columns, index_type='UNIQUE')
        self.assertEquals([f'Key name {too_long} must be <=64 characters long'], index.schema_errors)

    def test_missing_name_use_column(self):
        index = Index(name='', columns=['hey'], index_type='sup')
        self.assertEquals('hey', index.name)

    def test_missing_things(self):
        index = Index('blah', [], index_type='')
        self.assertEquals([
            'Missing index_type for index blah',
            'Missing columns for index blah',
        ], index.schema_errors)



from ..definitions.columns.column import Column
from ..definitions.table import Table
class RemoveColumn:
    """ Generates a partial SQL command to remove a column from a table """
    def __init__(self, column: Column):
        self._column = column

    def apply_to_table(self, table: Table):
        """ Removes the column from the table """
        table.remove_column(self.column)

    def __str__(self):
        return 'DROP `%s`' % self._column.name



import unittest
from .change_row import ChangeRow
from collections import OrderedDict
class ChangeRowTest(unittest.TestCase):
    def test_as_string(self):
        row = OrderedDict([('id', 10), ('age', '5'), ('dob', None)])
        change_row = ChangeRow('test_table', row)
        self.assertEquals("UPDATE `test_table` SET `age`='5', `dob`=NULL WHERE id='10';", str(change_row))



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_enum import TypeEnum
class TestTypeEnum(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = TypeEnum()
        returned = parser.parse("options enum('bob','joe','ray') not null default 'bob',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('options', parser._name)
        self.assertEquals('enum', parser._column_type)
        self.assertEquals(['bob', 'joe', 'ray'], parser._enum_values)
        self.assertFalse(parser._null)
        self.assertEquals('bob', parser._default)
        self.assertTrue(parser.has_comma)
        self.assertEquals(0, len(parser._parsing_errors))
        self.assertEquals(0, len(parser._parsing_warnings))

    def test_optional_default(self):

        # parse typical insert values
        parser = TypeEnum()
        returned = parser.parse("options enum('bob','joe','ray')")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('options', parser._name)
        self.assertEquals('enum', parser._column_type)
        self.assertEquals(['bob', 'joe', 'ray'], parser._enum_values)
        self.assertTrue(parser._null)
        self.assertEquals(None, parser._default)
        self.assertFalse(parser.has_comma)
        self.assertEquals(0, len(parser._parsing_errors))
        self.assertEquals(0, len(parser._parsing_warnings))

    def test_strip_backticks(self):

        # parse typical insert values
        parser = TypeEnum()
        returned = parser.parse("`options` enum('bob','joe','ray')")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertEquals('options', parser._name)

    def test_warning_for_no_quote_default(self):

        # parse typical insert values
        parser = TypeEnum()
        returned = parser.parse("options enum('bob','joe','ray') default bob")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertTrue('should have quotes' in parser._parsing_warnings[0])



import unittest
from .constraint import Constraint
class TestConstraint(unittest.TestCase):
    def test_string_conversion(self):
        constraint = Constraint(
            'user_id_fk',
            'user_id',
            'users',
            'id',
            on_delete='SET NULL',
            on_update='CASCADE',
        )

        self.assertEquals(
            'CONSTRAINT `user_id_fk` FOREIGN KEY (`user_id`) REFERENCES `users` (`id`) ON DELETE SET NULL ON UPDATE CASCADE',
            str(constraint)
        )
        self.assertEquals([], constraint.schema_errors)
        self.assertEquals([], constraint.schema_warnings)



class Database:
    _parsing_errors = None
    _schema_errors = None
    _parsing_warnings = None
    _schema_warnings = None
    _tables = None
    _rows = None
    _global_errors = None
    _global_warnings = None

    @property
    def tables(self):
        """ Public getter.  Returns a dict of table definitions, by table name

        :returns: A dict of table definitions, by table name
        :rtype: dict
        """
        return self._tables

    @property
    def rows(self):
        """ Public getter.  Returns a dict of row definitions, by as lists organized by table name

        :returns: A dict of row definitions, by table name
        :rtype: dict[list[Row]]
        """
        return self._rows

    @property
    def errors(self):
        """ Public getter.  Returns a list of errors

        :returns: A list of errors
        :rtype: list
        """
        global_errors = self._global_errors if self._global_errors is not None else []
        schema_errors = self.find_schema_errors()
        parsing_errors = self.find_parsing_errors()
        return [*global_errors, *parsing_errors, *schema_errors]

    @property
    def warnings(self):
        """ Public getter.  Returns a list of warnings

        :returns: A list of warnings
        :rtype: list[string]
        """
        global_warnings = self._global_warnings if self._global_warnings is not None else []
        schema_warnings = self.find_schema_warnings()
        parsing_warnings = self.find_parsing_warnings()
        return [*global_warnings, *parsing_warnings, *schema_warnings]

    def find_schema_warnings(self):
        """ Returns a list of human-readable warning messages (strings) that give any warnings about the schema itself.
        :return: List of warning messages
        :rtype: list
        """
        if self._schema_warnings is not None:
            return self._schema_warnings

        self._schema_warnings = [
            *self.find_all_row_schema_warnings(),
            *self.find_all_table_schema_warnings(),
        ]

        return self._schema_warnings

    def find_schema_errors(self):
        """ Returns a list of human-readable error messages (strings) that give any errors about the schema itself.
        :return: List of error messages
        :rtype: list
        """
        if self._schema_errors is not None:
            return self._schema_errors

        self._schema_errors = [
            *self.find_all_database_schema_errors(),
            *self.find_all_table_schema_errors(),
            *self.find_all_row_schema_errors(),
        ]

        return self._schema_errors

    def find_all_row_schema_errors(self):
        """ Returns a list of any schema error messages for the rows in this database

        :return: A list of schema error messages
        :rtype: list[string]
        """
        errors = []
        for (table_name, rows) in self.rows.items():
            for row in rows:
                if hasattr(row, 'schema_errors'):
                    errors.extend(row.schema_errors)

                # verify the columns which the inserts reference make sense
                if table_name not in self.tables:
                    errors.append(
                        f"Insert error: found an insert for table '{table_name}' but this table does not exist"
                    )

                elif row.num_explicit_columns:
                    table = self.tables[table_name]
                    for column_name in row.columns:
                        if column_name not in table.columns:
                            errors.append(
                                f"Insert error: insert command attempts to set column '{column_name}' for table '{table_name}' but the column does not exist in the table."
                            )
                else:
                    for row_values in row.raw_rows:
                        if len(row_values) != len(table.columns):
                            errors.append(
                                f"Insert error for table '{table_name}': insert command does not explicitly set column names and has a different number of values than the number of columns in the table"
                            )

        return errors

    def find_all_row_schema_warnings(self):
        """ Returns a list of any schema warning messages for the rows in this database

        :return: A list of schema warning messages
        :rtype: list[string]
        """
        warnings = []
        for rows in self.rows.values():
            for row in rows:
                if hasattr(row, 'schema_warnings'):
                    warnings.extend(row.schema_warnings)
        return warnings

    def find_all_table_schema_errors(self):
        """ Returns a list of any schema error messages for the tables in this database

        :return: A list of schema error messages
        :rtype: list[string]
        """
        errors = []
        for table in self.tables.values():
            if hasattr(table, 'schema_errors'):
                errors.extend(table.schema_errors)
        return errors

    def find_all_table_schema_warnings(self):
        """ Returns a list of any schema warnings messages for the tables in this database

        :return: A list of schema warnings messages
        :rtype: list[string]
        """
        warnings = []
        for table in self.tables.values():
            if hasattr(table, 'schema_warnings'):
                warnings.extend(table.schema_warnings)
        return warnings

    def find_all_database_schema_errors(self):
        """ Returns a list of any constraint error messages for this database

        :return: A list of constraint error messages
        :rtype: [string]
        """
        errors = []
        constraint_tables = {}
        for table in self.tables.values():
            if not table.constraints:
                continue

            for constraint in table.constraints.values():
                error = self.find_constraint_errors(table, constraint)
                if error:
                    errors.append(error)

                if constraint.name in constraint_tables:
                    errors.append(
                        f"Duplicate foreign key: foreign key named '{constraint.name}' exists in tables '{constraint_tables[constraint.name]}' and '{table.name}'"
                    )
                else:
                    constraint_tables[constraint.name] = table.name

        return errors

    def find_parsing_errors(self):
        """ Returns a list of human-readable error messages (strings) that give any errors from the parsing process
        :return: List of error messages
        :rtype: list
        """
        if self._parsing_errors is not None:
            return self._parsing_errors

        self._parsing_errors = [
            *self.find_all_table_parsing_errors(),
            *self.find_all_row_parsing_errors(),
        ]

        return self._parsing_errors

    def find_parsing_warnings(self):
        """ Returns a list of human-readable error messages (strings) that give any errors from the parsing process
        :return: List of error messages
        :rtype: list
        """
        if self._parsing_warnings is not None:
            return self._parsing_warnings

        self._parsing_warnings = [
            *self.find_all_table_parsing_warnings(),
            *self.find_all_row_parsing_warnings(),
        ]

        return self._parsing_warnings

    def find_all_table_parsing_errors(self):
        """ Returns a list of human-readable error messages (strings) from the table parsing process
        :return: List of error messages
        :rtype: list
        """
        errors = []
        for table in self.tables.values():
            if hasattr(table, 'parsing_errors'):
                errors.extend(table.parsing_errors)
        return errors

    def find_all_row_parsing_errors(self):
        """ Returns a list of human-readable error messages (strings) from the row parsing process
        :return: List of error messages
        :rtype: list
        """
        errors = []
        for (table_name, rows) in self.rows.items():
            for row in rows:
                if hasattr(row, 'parsing_errors'):
                    errors.extend(row.parsing_errors)
        return errors

    def find_all_table_parsing_warnings(self):
        """ Returns a list of human-readable warning messages (strings) from the table parsing process
        :return: List of warning messages
        :rtype: list
        """
        warnings = []
        for table in self.tables.values():
            if hasattr(table, 'parsing_warnings'):
                warnings.extend(table.parsing_warnings)
        return warnings

    def find_all_row_parsing_warnings(self):
        """ Returns a list of human-readable warning messages (strings) from the row parsing process
        :return: List of warning messages
        :rtype: list
        """
        warnings = []
        for (table_name, rows) in self.rows.items():
            for row in rows:
                if hasattr(table, 'parsing_warnings'):
                    warnings.extend(row.parsing_warnings)
        return warnings

    def find_constraint_errors(self, table, constraint):
        """ Returns None or a string with an error message for the given table and constraint

        :param table: The table being checked
        :param constraint: The constraint to check the table against
        :type table: mygrations.core.definitions.table
        :type constraint: mygrations.core.definitions.constraint
        :rtype: string|None
        """

        if constraint.foreign_table_name not in self.tables:
            return "Constraint error for foreign key `%s`: `%s`.`%s` references `%s`.`%s`, but table `%s` does not exist" % (
                constraint.name, table.name, constraint.column_name, constraint.foreign_table_name,
                constraint.foreign_column_name, constraint.foreign_table_name
            )
        foreign_table = self.tables[constraint.foreign_table_name]
        if constraint.foreign_column_name not in foreign_table.columns:
            return "Constraint error for foreign key `%s`: `%s`.`%s` references `%s`.`%s`, but column `%s`.`%s` does not exist" % (
                constraint.name, table.name, constraint.column_name, constraint.foreign_table_name,
                constraint.foreign_column_name, constraint.foreign_table_name, constraint.foreign_column_name
            )

        # the column exists but we may still have a constraint error.  That can happen in a few ways
        if constraint.column_name not in table.columns:
            return "Constraint error for foreign key `%s`: sets constraint on column `%s`.`%s`, but this column does not exist" % (
                constraint.name, table.name, constraint.column_name
            )
        table_column = table.columns[constraint.column_name]
        foreign_column = foreign_table.columns[constraint.foreign_column_name]

        if not table.column_is_indexed(table.columns[table_column.name]):
            return "Constraint error for foreign key `%s`: missing index. `%s`.`%s` does not have an index and therefore cannot be used in a foreign key constraint" % (
                constraint.name, table.name, table_column.name
            )

        # we have a few attributes that must must match exactly and have easy-to-produce errors
        for attr in ['column_type', 'length', 'character_set', 'collate']:
            table_value = getattr(table_column, attr)
            foreign_value = getattr(foreign_column, attr)
            if table_value != foreign_value:
                return "Constraint error for foreign key `%s`: %s mismatch. `%s`.`%s` is '%s' but `%s`.`%s` is '%s'" % (
                    constraint.name, attr.replace('_', ' '), table.name, constraint.column_name, table_value,
                    foreign_table.name, foreign_column.name, foreign_value
                )

        # unsigned are separate because they get a slightly different message
        if table_column.unsigned and not foreign_column.unsigned:
            return "Constraint error for foreign key `%s`: unsigned mistmatch. `%s`.`%s` is unsigned but `%s`.`%s` is not" % (
                constraint.name, table.name, table_column.name, foreign_table.name, foreign_column.name
            )

        if not table_column.unsigned and foreign_column.unsigned:
            return "Constraint error for foreign key `%s`: unsigned mistmatch. `%s`.`%s` is unsigned but `%s`.`%s` is not" % (
                constraint.name, foreign_table.name, foreign_column.name, table.name, table_column.name
            )

        # if the constraint has a SET NULL but the column cannot be null, then 1215
        if (constraint.on_delete == 'SET NULL' or constraint.on_update == 'SET NULL') and not table_column.null:
            message_parts = []
            if constraint.on_delete == 'SET NULL':
                message_parts.append('ON DELETE')
            if constraint.on_update == 'SET NULL':
                message_parts.append('ON UPDATE')
            return "Constraint error for foreign key `%s`: invalid SET NULL. `%s`.`%s` is not allowed to be null but the foreign key attempts to set the value to null %s" % (
                constraint.name, table.name, table_column.name, ' and '.join(message_parts)
            )

        # if the column the constraint is on doesn't have an index, then 1215
        if not foreign_table.column_is_indexed(foreign_column):
            return "Constraint error for foreign key `%s`: missing index. `%s`.`%s` references `%s`.`%s` but `%s`.`%s` does not have an index and therefore cannot be used in a foreign key constraint" % (
                constraint.name, table.name, table_column.name, foreign_table.name, foreign_column.name,
                foreign_table.name, foreign_column.name
            )

        return None

    def unfulfilled_fks(self, table):
        """ Returns a dictionary with information about all constraints in the table which are not fulfilled by this database

        If all foreign keys are fulfilled by the database structure then an empty dict is returned

        The returned dictionary contains a key with the name of every foreign key that cannot
        be fulfilled.  The value in the dictionary will be another dictionary containing
        'error' (an error message stating exactly what the problem is) and 'foreign_key'
        (the actual foreign key definition that cannot be fulfilled).

        This is similar to find_all_database_schema_errors but is used in a specific way by the migration process,
        so gets its own method (although both rely heavily on the `find_constraint_errors` method.

        :param table: The table to check
        :type table: mygrations.formats.mysql.definitions.table
        :return: Dictionary with information on all foreign keys that cannot be fulfilled
        :rtype: dict
        """
        if not table.constraints:
            return {}

        unfulfilled = {}
        for (constraint_name, constraint) in table.constraints.items():
            error = self.find_constraint_errors(table, constraint)
            if error:
                unfulfilled[constraint_name] = {"error": error, "foreign_key": constraint}

        return unfulfilled

    def add_table(self, table):
        """ Adds a table to the database

        :param table: The table to add
        :type table: mygrations.formats.mysql.definitions.table
        """
        if table.name in self._tables:
            raise ValueError('Cannot add table %s to database because it already exists' % table.name)

        self._schema_errors = None
        self._schema_warnings = None
        self._tables[table.name] = table

    def remove_table(self, table):
        """ Removes a table from the database

        :param table: The table to remove
        :type table: mygrations.formats.mysql.definitions.table
        """
        if not table.name in self._tables:
            raise ValueError('Cannot remove table %s from database because it does not exist' % table.name)

        self._schema_errors = None
        self._schema_warnings = None
        self._tables.pop(table.name, None)

    def apply_operation(self, table_name, operation):
        """ Updates the database object according to the operation

        :param table_name: The table that the operation is being applied to
        :param operation: The operation to apply
        :type table: string|mygrations.formats.mysql.definitions.table
        :type operation: mygrations.formats.mysql.mygration.operations.*
        """
        if type(table_name) != str:
            table_name = table_name.name

        if not table_name in self._tables:
            raise ValueError('Cannot apply operation to table %s because that table does not exist' % table_name)

        # the table applies the operation
        self._schema_errors = None
        self._schema_warnings = None
        self._tables[table_name].apply_operation(operation)

    def apply_to_source(self, operation):
        """ Updates the actual source of the database object: usually the database or a .sql file

        This method and :meth:`mygrations.formats.mysql.defintions.database.apply_operation` sound very similar.
        The difference is that the latter operates on the in-memory structure store only, and makes
        no actual changes.  Instead it is primarily used for internal purposes.  This method actually
        makes changes.

        :param operation: The operation to apply
        :type operation: mygrations.formats.mysql.mygration.operations.*
        """
        raise NotImplementedError("Someone forgot to finish something")



from __future__ import annotations
from typing import List
class Base:
    _parsing_errors = None
    _parsing_warnings = None

    @property
    def parsing_errors(self) -> List[str]:
        """ Public getter.  Returns a list of parsing errors """
        return self._parsing_errors if self._parsing_errors is not None else []

    @property
    def parsing_warnings(self) -> List[str]:
        """ Public getter.  Returns a list of parsing errors """
        return self._parsing_warnings if self._parsing_warnings is not None else []



import os, os.path
from mygrations.helpers.dotenv import dotenv
from mygrations.helpers.db_credentials import DbCredentials
from mygrations.drivers.pymysql.pymysql import PyMySQL
class Base(object):
    options = None
    config = None
    _config_loaded = False

    def __init__(self, options):
        self.options = options

    def get_driver(self):
        if 'driver' not in self.options:
            if 'connection' in self.options:
                self.options['driver'] = PyMySQL(connection=self.options['connection'])
            else:
                self._load_config()
                self.options['driver'] = PyMySQL(DbCredentials(self.config['env_file'], self.config))
        return self.options['driver']

    def get_sql_files(self):
        if 'sql_files' not in self.options:
            self._load_config()
            self.options['sql_files'] = self.config['files_directory']
        return self.options['sql_files']

    def _load_config(self):
        if self._config_loaded:
            return True

        if 'config' not in self.options:
            raise ValueError(f"Missing required 'config' parameter in options for commands '{self.__class__.__name__}'")

        # load up the mygration configuration (which includes the path to the files we will import)
        # note that self.options['config'] is supposed to contain the filename or absolute path to the
        # mygrations configuration file.  By default, this is called `mygrate.conf`.  In return
        # we get the absolute path to this file, which contains the config we need to do anything from
        # the command line (we shouldn't end up here if called programmatically).

        # this can be slightly confusing because of the mixture of self.config and self.options.  It's a simple
        # difference though: self.options is a dictionary passed in by whoever calls us (probably the command line runner),
        # while self.config is a dictionary containing configuration settings loaded out of the mygrate.conf file.
        config_abs_path = self._resolve_and_check_config_file_path(self.options['config'])
        self.config = dotenv(config_abs_path)
        if 'files_directory' not in self.config:
            raise ValueError(
                "Configuration file is missing 'files_directory' configuration, which should point to the directory containing your *.sql files"
            )
        self.config['files_directory'] = self._relative_to_config_abs_path(
            self.config['files_directory'], config_abs_path
        )

        if 'env_file' not in self.config:
            raise ValueError(
                "Configuration file is missing 'env_file' configuration, which should point to the .env file for your application (so mygrations can fetch database credentials"
            )
        # convert the env file path to an absolute path
        self.config['env_file'] = self._relative_to_config_abs_path(self.config['env_file'], config_abs_path)
        self._config_loaded = True

    def execute(self):
        raise NotImplementedError()

    def _resolve_and_check_config_file_path(self, file_path):
        if not len(file_path):
            raise ValueError('Missing path to configuration file')

        # We want to find the location of the mygrate.conf file.  For the most part expect file_path to literally
        # be `mygrate.conf`, as that is the default sent in by the runner.  However the user could have specified
        # something else.  If we have an absolute path specified then that is the only thing we want to check,
        # and we want to throw an exception if it does not match.  However, if it is not an absolute path we
        # want to search for the given file in current directory or its parents.
        if file_path[0] == '/':
            if not os.path.isfile(file_path):
                raise ValueError("Specified configuration file '%s' does not exist" % file_path)
            return file_path

        directories = ('%s/%s' % (os.getcwd(), file_path)).strip('/').split('/')
        filename = directories.pop()
        ndirectories = len(directories)
        for i in range(ndirectories):
            abs_path = '/%s/%s' % ('/'.join(directories[:ndirectories - i]), filename)
            if os.path.isfile(abs_path):
                return abs_path

        raise ValueError("Could not find configuration file '%s' in current directory or parents" % file_path)

    def _relative_to_config_abs_path(self, relative_path, config_abs_path):
        # Our file directory and our .env file path should both be relative to the location of the mygrate.conf file.
        # The exact path to the mygrate.conf file can change because it is allowed to be in a parent of the
        # current working directory.  Therefore we need a way to take a relative path and turn it into an absolute
        # path using the absolute path to the configuration file (i.e. the mygrate.conf file).
        # This method does just that.
        # However, don't muck with a path if it already absolute
        if relative_path[0] == '/':
            return relative_path

        return '%s/%s' % (os.path.dirname(config_abs_path), relative_path)



import unittest

from mygrations.core.parse.rule_literal import RuleLiteral
class RuleLiteralTest(unittest.TestCase):
    def get_rule(self, name, literal):

        return RuleLiteral(False, {'name': name, 'value': literal}, {})

    def test_name_not_required(self):

        rule = self.get_rule('', 'ASDF')
        self.assertEquals(rule.name, 'ASDF')

    def test_value_required(self):

        with self.assertRaises(ValueError):

            self.get_rule('bob', '')

    def test_can_init_with_name_and_value(self):

        rule = self.get_rule('bob', ',')
        self.assertEquals(rule.name, 'bob')
        self.assertEquals(rule.literal, ',')

    def test_match_beginning_only(self):

        rule = self.get_rule('bob', ',')
        self.assertFalse(rule.parse('hey,'))
        self.assertEquals('', rule.result)

    def test_leftovers_is_input_for_no_match(self):

        rule = self.get_rule('bob', ',')
        string = 'hey,'
        rule.parse(string)

        self.assertEquals(string, rule.leftovers)

    def test_no_leftovers_for_full_match(self):

        string = '23483438'
        rule = self.get_rule('bob', string)

        self.assertTrue(rule.parse(string))
        self.assertEquals(string, rule.result)
        self.assertEquals('', rule.leftovers)

    def test_calc_leftovers_trim(self):

        rule = self.get_rule('bob', ',')

        string = ', bob'

        self.assertTrue(rule.parse(string))
        self.assertEquals(',', rule.result)
        self.assertEquals('bob', rule.leftovers)



import unittest
from .constraint import Constraint
class TestConstraint(unittest.TestCase):
    def test_can_create(self):
        constraint = Constraint(
            'user_id_fk',
            'user_id',
            'users',
            'id',
        )

        self.assertEquals('user_id_fk', constraint.name)
        self.assertEquals('user_id', constraint.column_name)
        self.assertEquals('users', constraint.foreign_table_name)
        self.assertEquals('id', constraint.foreign_column_name)
        self.assertEquals('RESTRICT', constraint.on_delete)
        self.assertEquals('RESTRICT', constraint.on_update)
        self.assertEquals([], constraint.schema_errors)

    def test_action(self):
        constraint = Constraint(
            'user_id_fk',
            'user_id',
            'users',
            'id',
            on_delete='GARBAGE',
            on_update='whatever',
        )
        self.assertEquals([
            "ON DELETE action of 'GARBAGE' for constraint 'user_id_fk' is not a valid ON DELETE action",
            "ON UPDATE action of 'WHATEVER' for constraint 'user_id_fk' is not a valid ON UPDATE action",
        ], constraint.schema_errors)



import unittest
from .enum import Enum
class TestEnum(unittest.TestCase):
    def test_string_conversion(self):
        column = Enum('test_column', 'ENUM', enum_values=['asdf', 'qwerty'], default='qwerty')

        self.assertEquals("`test_column` ENUM('asdf', 'qwerty') DEFAULT \'qwerty\'", str(column))
        self.assertEquals([], column.schema_errors)
        self.assertEquals([], column.schema_warnings)

    def test_default_errors(self):
        column = Enum('test_column', 'ENUM', enum_values=['asdf', 'qwerty'], default='adf')
        self.assertEquals(["Default value for 'ENUM' column 'test_column' is not in the list of allowed values"],
                          column.schema_errors)

        self.assertEquals([], Enum('test_column', 'ENUM', enum_values=['asdf', 'qwerty']).schema_errors)



import unittest

from mygrations.helpers.dotenv import dotenv, DotEnvSyntaxError
import io
import tempfile
class DotEnvTest(unittest.TestCase):

    dotenv = None
    test_string = 'test string'

    def setUp(self):

        self.dotenv = dotenv()

    # get_contents should accept a number of parameters.
    # It should accept a stringIO wrapper
    def test_get_contents_stringIO(self):

        self.assertEquals(self.test_string, self.dotenv.get_contents(io.StringIO(self.test_string)))

    # it should also accept an actual string
    def test_get_contents_string(self):

        self.assertEquals(self.test_string, self.dotenv.get_contents(self.test_string))

    # as well as a more general file pointer
    def test_get_contents_fp(self):

        fp = tempfile.TemporaryFile()
        fp.write(self.test_string.encode(encoding='UTF-8'))
        fp.seek(0)
        self.assertEquals(self.test_string, self.dotenv.get_contents(fp))
        fp.close()

    # it should also accept a filename
    def test_get_contents_filename(self):

        filename = '%s/unit_mygrations_dotenv' % tempfile.gettempdir()
        fp = open(filename, 'w')
        fp.write(self.test_string)
        fp.close()

        self.assertEquals(self.test_string, self.dotenv.get_contents(filename))

    def test_json(self):

        # it should detect json and parse it as such
        parsed = self.dotenv.parse('{ "key": "value",\n "name": "bob" }')

        self.assertEquals({'key': 'value', 'name': 'bob'}, parsed)

    def test_dont_interrupt_syntax_error(self):

        # a syntax error from parse_line should not be stopped
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse, 'name')

    def test_parse_each_line(self):

        # otherwise each line should be parsed
        parsed = self.dotenv.parse('key = value\nname = bob #sup')

        self.assertEquals({'key': 'value', 'name': 'bob'}, parsed)

    def test_parse_skip_empty(self):

        # some whitespace checks and a few parts just because
        parsed = self.dotenv.parse('key = value\n\n\n\n   name = "bob" # sup\n another = "test\\"value"#okay')

        self.assertEquals({'key': 'value', 'name': 'bob', 'another': 'test"value'}, parsed)

    def test_opening_comment(self):

        self.assertEquals([False, False], self.dotenv.parse_line('# name = test', '#'))

    def test_opening_comment_with_space(self):

        # white space is stripped first, so it still starts with a comment if it has spaces
        self.assertEquals([False, False], self.dotenv.parse_line(' # name = test', '#'))

    def test_any_comment_charater(self):

        # our comment character should be respected
        self.assertEquals([False, False], self.dotenv.parse_line('; name = test', ';'))

    def test_multi_character_comment(self):

        # and it is allowed to be more than one character long
        self.assertEquals([False, False], self.dotenv.parse_line('weirdcomment name = test', 'weirdcomment'))

    def test_ignore_empty_lines(self):

        # empty lines (which includes lines with all white space) are ignored
        self.assertEquals([False, False], self.dotenv.parse_line('  ', '#'))

    def test_invalid_format(self):

        # supported format is something like name = value or name: value
        # if neither of the separators is present then we should get a syntax error
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, 'sup', '#')

    def test_empty_value_equal(self):

        # empty values are allowed
        self.assertEquals(['key', ''], self.dotenv.parse_line('key=', '#'))

    def test_empty_value_colon(self):

        # colons act exactly the same way
        self.assertEquals(['key', ''], self.dotenv.parse_line('key:', '#'))

    def test_empty_value_equal_spaces(self):

        # and spaces are ignored
        self.assertEquals(['key', ''], self.dotenv.parse_line(" key \t=\t  ", '#'))

    def test_empty_value_equal_colon(self):

        # and spaces are ignored
        self.assertEquals(['key', ''], self.dotenv.parse_line(" key \t:\t  ", '#'))

    def test_value_no_quotes_equal(self):

        # a value without quotes should be easy
        self.assertEquals(['key', 'asdf'], self.dotenv.parse_line("key=asdf", '#'))

    def test_value_no_quotes_colon(self):

        # a value without quotes should be easy
        self.assertEquals(['key', 'asdf'], self.dotenv.parse_line("key:asdf", '#'))

    def test_value_no_quotes_equal_spaces(self):

        # spaces are still ignored at the beginning/end of a part
        self.assertEquals(['key', 'asdf bob'], self.dotenv.parse_line("key= asdf bob  \t", '#'))

    def test_value_no_quotes_equal_colon(self):

        # spaces are still ignored at the beginning/end of a part
        self.assertEquals(['key', 'asdf bob'], self.dotenv.parse_line("key : asdf bob  \t", '#'))

    def test_value_no_quotes_equal_comment(self):

        # and comments at the end are ignored (including spaces before comments)
        self.assertEquals(['key', 'asdf bob'], self.dotenv.parse_line("key = asdf bob  \t# a comment", '#'))

    def test_value_no_quotes_equal_colon(self):

        # and comments at the end are ignored (including spaces before comments)
        self.assertEquals(['key', 'asdf bob'], self.dotenv.parse_line("key : asdf bob  \t; a comment", ';'))

    def test_no_lone_quotes_double(self):

        # a quote character inside the value by itself is invalid
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, 'name = valu"e', '#')

    def test_no_lone_quotes_single(self):

        # a quote character inside the value by itself is invalid
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, "name = valu'e", '#')

    def test_empty_single(self):

        # easy
        self.assertEquals(['db', ''], self.dotenv.parse_line("db = ''", '#'))

    def test_empty_double(self):

        # easy
        self.assertEquals(['db', ''], self.dotenv.parse_line('db = ""', '#'))

    def test_empty_single_comment(self):

        # easy
        self.assertEquals(['db', ''], self.dotenv.parse_line("db = '' # database name", '#'))

    def test_empty_double_comment(self):

        # easy
        self.assertEquals(['db', ''], self.dotenv.parse_line('db = "" ; database name', ';'))

    def test_single_quotes(self):

        # I shouldn't get the quotes back.  Also white space should
        # be ignored still
        self.assertEquals(['db', 'dijere'], self.dotenv.parse_line("db = 'dijere' ", '#'))

    def test_double_quotes(self):

        # I shouldn't get the quotes back.  Also white space should
        # be ignored still
        self.assertEquals(['db', 'dijere'], self.dotenv.parse_line('db = "dijere" ', '#'))

    def test_no_closing_quote_single(self):

        # syntax error if we have an opening quote with no close
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, "name = 'test", '#')

    def test_no_closing_quote_double(self):

        # syntax error if we have an opening quote with no close
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, 'name = "test', '#')

    def test_double_quotes_with_comment(self):

        # comments after the quotes should be ignored (along with whitespace)
        self.assertEquals(['db', 'bob'], self.dotenv.parse_line('db = "bob" ; database name', ';'))

    def test_single_quotes_with_comment(self):

        # comments after the quotes should be ignored (along with whitespace)
        self.assertEquals(['db', 'bob'], self.dotenv.parse_line("db = 'bob' \t# database name", '#'))

    def test_text_outside_of_double(self):

        # anything outside of the quotes and before a comment results in a syntax error
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, 'name = "test" sup # hey', '#')

    def test_text_outside_of_single(self):

        # anything outside of the quotes and before a comment results in a syntax error
        self.assertRaises(DotEnvSyntaxError, self.dotenv.parse_line, "name = 'test' sup ; hey", ';')

    def test_allow_ending_semicolon(self):

        # anything outside of the quotes and before a comment results in a syntax error
        # except allow to end the line with a semi-colon.
        self.assertEquals(['name', 'test'], self.dotenv.parse_line("name = 'test';", '#'))

    def test_escaped_single_quote(self):

        # quotes can be escaped inside quotes
        self.assertEquals(['db', "asdf'qwerty'"], self.dotenv.parse_line("db = 'asdf\\'qwerty\\''", '#'))

    def test_escaped_double_quote(self):

        # quotes can be escaped inside quotes
        self.assertEquals(['db', 'asdf"qwerty"'], self.dotenv.parse_line('db = "asdf\\"qwerty\\""', '#'))

    def test_preserve_other_slashes(self):

        # other slashes are left alone
        self.assertEquals(['db', 'asdf\\bob'], self.dotenv.parse_line('db = "asdf\\bob"', '#'))

    def test_double_quote_in_single_quote(self):

        # double quote inside single quotes are just regular characters
        self.assertEquals(['db', 'asdf"bob'], self.dotenv.parse_line("db = 'asdf\"bob'", '#'))

    def test_single_quote_in_double_quote(self):

        # single quote inside double quotes are just regular characters
        self.assertEquals(['db', "asdf'bob"], self.dotenv.parse_line('db = "asdf\'bob"', '#'))



from ..definitions.index import Index
from ..definitions.table import Table
class AddIndex:
    """ Generates a partial SQL command to add an index to a table """
    _index: Index = None

    def __init__(self, index: Index):
        self._index = index

    def apply_to_table(self, table: Table):
        """ Adds the index to the table """
        table.add_index(self._index)

    def __str__(self):
        return 'ADD %s' % str(self._index)



import unittest
from .remove_index import RemoveIndex
from ..definitions.index import Index
class RemoveIndexTest(unittest.TestCase):
    def test_as_string(self):
        index = Index('user_id', ['user_id', 'account_id'], 'index')
        self.assertEquals("DROP KEY `user_id`", str(RemoveIndex(index)))

        index = Index('user_id', ['user_id'], 'primary')
        self.assertEquals("DROP PRIMARY KEY", str(RemoveIndex(index)))



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_numeric import TypeNumeric
class TestTypeNumeric(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = TypeNumeric()
        returned = parser.parse("created int(10) not null default 0,")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('created', parser._name)
        self.assertEquals('int', parser._column_type)
        self.assertEquals('10', parser._length)
        self.assertFalse(parser._unsigned)
        self.assertFalse(parser._null)
        self.assertFalse(parser._auto_increment)
        self.assertEquals('0', parser._default)
        self.assertTrue(parser.has_comma)
        self.assertEquals(0, len(parser._parsing_errors))
        self.assertEquals(0, len(parser._parsing_warnings))

    def test_auto_increment(self):

        # parse typical insert values
        parser = TypeNumeric()
        returned = parser.parse("created int(10) not null auto_increment,")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('created', parser._name)
        self.assertEquals('int', parser._column_type)
        self.assertEquals('10', parser._length)
        self.assertTrue(parser._auto_increment)
        self.assertFalse(parser._unsigned)
        self.assertFalse(parser._null)
        self.assertEquals(None, parser._default)
        self.assertTrue(parser.has_comma)
        self.assertEquals(0, len(parser._parsing_errors))
        self.assertEquals(0, len(parser._parsing_warnings))

    def test_optional_default(self):

        # parse typical insert values
        parser = TypeNumeric()
        returned = parser.parse("created int(10) UNSIGNED")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('created', parser._name)
        self.assertEquals('int', parser._column_type)
        self.assertTrue(parser._unsigned)
        self.assertEquals('10', parser._length)
        self.assertTrue(parser._null)
        self.assertEquals(None, parser._default)
        self.assertFalse(parser.has_comma)
        self.assertEquals(0, len(parser._parsing_errors))
        self.assertEquals(0, len(parser._parsing_warnings))

    def test_strip_backticks(self):

        # parse typical insert values
        parser = TypeNumeric()
        returned = parser.parse("`created` int(10) UNSIGNED")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)
        self.assertEquals('created', parser._name)



import unittest

from mygrations.formats.mysql.file_reader.parsers.type_datetime import TypeDateTime
class TestTypeDateTime(unittest.TestCase):
    def test_simple(self):

        # parse typical insert values
        parser = TypeDateTime()
        returned = parser.parse("name datetime not null default 'YY-MM-DD HH-MM-SS',")

        self.assertTrue(parser.matched)
        self.assertEquals('', returned)

        self.assertEquals('name', parser._name)
        self.assertEquals('datetime', parser._column_type)
        self.assertFalse(parser._null)
        self.assertEquals('YY-MM-DD HH-MM-SS', parser._default)
        self.assertTrue(parser.has_comma)
        self.assertTrue(parser._character_set is None)
        self.assertTrue(parser._collate is None)



from mygrations.formats.mysql.definitions import columns
from mygrations.core.definitions.columns.column import Column
class Type:
    column_type_map = {
        'INTEGER': columns.Numeric,
        'INT': columns.Numeric,
        'SMALLINT': columns.Numeric,
        'TINYINT': columns.Numeric,
        'MEDIUMINT': columns.Numeric,
        'BIGINT': columns.Numeric,
        'DECIMAL': columns.Numeric,
        'NUMERIC': columns.Numeric,
        'FLOAT': columns.Numeric,
        'DOUBLE': columns.Numeric,
        'BIT': columns.Numeric,
        'DATE': columns.Date,
        'DATETIME': columns.Date,
        'TIMESTAMP': columns.Date,
        'TIME': columns.Date,
        'YEAR': columns.Date,
        'CHAR': columns.String,
        'VARCHAR': columns.String,
        'BINARY': columns.String,
        'VARBINARY': columns.String,
        'TINYBLOB': columns.String,
        'BLOB': columns.String,
        'MEDIUMBLOB': columns.String,
        'LONGBLOB': columns.String,
        'TINYTEXT': columns.String,
        'TEXT': columns.String,
        'MEDIUMTEXT': columns.String,
        'LONGTEXT': columns.String,
        'JSON': columns.String,
        'ENUM': columns.Enum,
        'SET': columns.Enum,
    }

    def as_definition(self) -> Column:
        """
        Returns the parsed column as a MySQL Column Definition

        This is here because a single parser class can end up working for columns of completely different
        types.  Therefore, we can't put the column definition directly in our inheritence tree and have
        to build and return a new one based on the type we have.
        """
        column_type = self._column_type.upper()
        if column_type not in self.column_type_map:
            raise ValueError(
                f"Found an unknown column type, '{column_type}' that does not have a corresponding column class"
            )
        column_class = self.column_type_map[column_type]
        return column_class(
            name=self._name,
            column_type=column_type,
            length=getattr(self, '_length', None),
            null=getattr(self, '_null', None),
            has_default=getattr(self, '_has_default', False),
            default=getattr(self, '_default', None),
            unsigned=getattr(self, '_unsigned', None),
            character_set=getattr(self, '_character_set', None),
            collate=getattr(self, '_collate', None),
            auto_increment=getattr(self, '_auto_increment', None),
            enum_values=getattr(self, 'enum_values', None),
            parsing_errors=getattr(self, '_parsing_errors', None),
            parsing_warnings=getattr(self, '_parsing_warnings', None),
        )
